"""Module for Variation Normalization."""
from typing import Optional, List, Tuple, Dict
from urllib.parse import quote

from ga4gh.vrsatile.pydantic.vrs_models import Text
from ga4gh.vrsatile.pydantic.vrsatile_models import VariationDescriptor, \
    GeneDescriptor
from ga4gh.vrs import models
from ga4gh.core import ga4gh_identify
from gene.query import QueryHandler as GeneQueryHandler
from uta_tools.data_sources import SeqRepoAccess, UTADatabase
from uta_tools.schemas import ResidueMode

from variation import logger
from variation.schemas.token_response_schema import GeneMatchToken, Token
from variation.schemas.validation_response_schema import ValidationSummary, \
    ValidationResult


class Normalize:
    """The Normalize class used to normalize a given variation."""

    def __init__(self, seqrepo_access: SeqRepoAccess, uta: UTADatabase,
                 gene_normalizer: GeneQueryHandler) -> None:
        """Initialize Normalize class.

        :param SeqRepoAccess seqrepo_access: Access to SeqRepo data queries
        :param UTADatabase uta: Access to UTA database and queries
        :parm QueryHandler gene_normalizer: Access to gene-normalizer queries
        """
        self.seqrepo_access = seqrepo_access
        self.uta = uta
        self.warnings = list()
        self._gene_norm_cache = dict()
        self.gene_normalizer = gene_normalizer

    @staticmethod
    def get_valid_result(q: str, validations: ValidationSummary,
                         warnings: List) -> ValidationResult:
        """Get valid result from ValidationSummary

        :param str q: Query string
        :param ValidationSummary validations: Validation summary for query
        :param List warnings: List of warnings
        :return: Valid Validation Result
        """
        # For now, only use first valid result
        valid_result = None
        if validations and validations.valid_results:
            for r in validations.valid_results:
                if r.is_mane_transcript and r.variation:
                    valid_result = r
                    break
        if not valid_result:
            warning = f"Unable to find MANE Transcript for {q}."
            logger.warning(warning)
            warnings.append(warning)
            if validations and validations.valid_results:
                valid_result = validations.valid_results[0]
        return valid_result

    def normalize(self, q: str, validations: ValidationSummary,
                  warnings: List) -> Optional[VariationDescriptor]:
        """Normalize a given variation.

        :param str q: The variation to normalize
        :param ValidationSummary validations: Invalid and valid results
        :param List warnings: List of warnings
        :return: An variation descriptor for a valid result if one exists.
            Else, None.
        """
        if not q:
            resp, warnings = self.no_variation_entered()
        else:
            label = q.strip()
            _id = f"normalize.variation:{quote(' '.join(label.split()))}"
            if len(validations.valid_results) > 0:
                valid_result = self.get_valid_result(q, validations, warnings)
                resp, warnings = self.get_variation_descriptor(
                    label, valid_result.variation, valid_result, _id, warnings)
            else:
                if not label:
                    resp, warnings = self.no_variation_entered()
                else:
                    resp, warnings = self.text_variation_resp(label, _id, warnings)
        self.warnings = warnings
        return resp

    @staticmethod
    def text_variation_resp(
            label: str, _id: str,
            warnings: List) -> Tuple[VariationDescriptor, List]:
        """Return text variation for queries that could not be normalized

        :param str label: Initial input query
        :param str _id: _id field for variation descriptor
        :param List warnings: List of warnings
        :return: Variation descriptor, warnings
        """
        warning = f"Unable to normalize {label}"
        text = models.Text(definition=label, type="Text")
        text._id = ga4gh_identify(text)
        resp = VariationDescriptor(
            id=_id,
            label=label,
            variation=Text(**text.as_dict())
        )
        if not warnings:
            warnings.append(warning)
            logger.warning(warning)
        return resp, warnings

    def get_variation_descriptor(
            self, label: str, variation: Dict, valid_result: ValidationResult,
            _id: str, warnings: List, gene: Optional[str] = None
    ) -> Tuple[VariationDescriptor, List]:
        """Return variation descriptor and warnings

        :param str label: Initial input query
        :param Dict variation: VRS variation object
        :param ValidationResult valid_result: Valid result for query
        :param str _id: _id field for variation descriptor
        :param List warnings: List of warnings
        :param Optional[str] gene: Gene symbol
        :return: Variation descriptor, warnings
        """
        variation_id = variation["_id"]
        identifier = valid_result.identifier
        token_type = \
            valid_result.classification_token.token_type.lower()

        vrs_ref_allele_seq = None
        if "uncertain" in token_type:
            warnings = ["Ambiguous regions cannot be normalized"]
        elif "range" not in token_type:
            if variation["type"] == "Allele":
                vrs_ref_allele_seq = self.get_ref_allele_seq(
                    variation, identifier
                )
            elif variation["type"] == "AbsoluteCopyNumber":
                vrs_ref_allele_seq = self.get_ref_allele_seq(
                    variation["subject"], identifier
                )

        if valid_result.gene_tokens:
            gene_token = valid_result.gene_tokens[0]
            gene_context = self.get_gene_descriptor(gene_token=gene_token)
        else:
            if gene:
                gene_context = self.get_gene_descriptor(gene=gene)
            else:
                gene_context = None

        return VariationDescriptor(
            id=_id,
            label=label,
            variation_id=variation_id,
            variation=variation,
            molecule_context=valid_result.classification_token.molecule_context,
            structural_type=valid_result.classification_token.so_id,
            vrs_ref_allele_seq=vrs_ref_allele_seq if vrs_ref_allele_seq else None,
            gene_context=gene_context,
            type="VariationDescriptor"
        ), warnings

    @staticmethod
    def no_variation_entered() -> Tuple[None, List[str]]:
        """Return response when no variation queried.

        :return: None, list of warnings
        """
        warnings = ["No variation was entered to normalize"]
        logger.warning(warnings)
        return None, warnings

    def get_gene_descriptor(
            self, gene_token: Optional[GeneMatchToken] = None,
            gene: Optional[str] = None) -> Optional[GeneDescriptor]:
        """Return a GA4GH Gene Descriptor using Gene Normalization.

        :param Optional[GeneMatchToken] gene_token: A gene token
        :param Optional[str] gene: Gene query
        :return: A gene descriptor for a given gene if a record exists in
            gene-normalizer.
        """
        if gene_token:
            gene_query = gene_token.matched_value
        elif gene:
            gene_query = gene
        else:
            return None

        if gene_query in self._gene_norm_cache:
            return self._gene_norm_cache[gene_query]
        else:
            response = self.gene_normalizer.normalize(gene_query)
            if response.gene_descriptor:
                gene_descriptor = response.gene_descriptor
                self._gene_norm_cache[gene_query] = gene_descriptor
                return gene_descriptor
            return None

    def get_ref_allele_seq(self, allele: Dict,
                           identifier: str) -> Optional[str]:
        """Return ref allele seq for transcript.

        :param Dict allele: VRS Allele object
        :param str identifier: Identifier for allele
        :return: Ref seq allele
        """
        start = None
        end = None
        interval = allele["location"]["interval"]
        ival_type = interval["type"]
        if ival_type == "SequenceInterval":
            if interval["start"]["type"] == "Number":
                start = interval["start"]["value"]
                end = interval["end"]["value"]

                if start == end:
                    return None

        if start is None and end is None:
            return None

        ref, _ = self.seqrepo_access.get_reference_sequence(
            identifier, start, end, residue_mode=ResidueMode.INTER_RESIDUE)

        return ref

    def _is_token_type(self, valid_result_tokens: List,
                       token_type: str) -> bool:
        """Return whether or not token_type is in valid_result_tokens.

        :param List valid_result_tokens: Valid token matches
        :param str token_type: The token's type
        :return: Whether or not token_type is in valid_result_tokens
        """
        for t in valid_result_tokens:
            if t.token_type == token_type:
                return True
        return False

    def _get_instance_type_token(self, valid_result_tokens: List,
                                 instance_type: Token) -> Optional[Token]:
        """Return the tokens for a given instance type.

        :param List valid_result_tokens: A list of valid tokens for the input
            string
        :param Token instance_type: The instance type to check
        :return: Token for a given instance type
        """
        for t in valid_result_tokens:
            if isinstance(t, instance_type):
                return t
        return None



"""The module for Silent Mutation Validation."""
from variation.schemas.classification_response_schema import \
    ClassificationType
from variation.schemas.token_response_schema import Token
from .polypeptide_sequence_variation_base import PolypeptideSequenceVariationBase


class SilentMutation(PolypeptideSequenceVariationBase):
    """The Silent Mutation Validator class."""

    def variation_name(self) -> str:
        """Return the variation name."""
        return "silent mutation"

    def is_token_instance(self, t: Token) -> bool:
        """Check that token is Silent Mutation."""
        return t.token_type == "SilentMutation"

    def validates_classification_type(
            self, classification_type: ClassificationType) -> bool:
        """Return whether or not the classification type is silent
        mutation.
        """
        return classification_type == ClassificationType.SILENT_MUTATION



"""A module for the Genomic Deletion Range Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class GenomicDeletionRangeClassifier(SetBasedClassifier):
    """The Genomic Deletion Range Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Genomic Deletion Range classification type."""
        return ClassificationType.GENOMIC_DELETION_RANGE

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["Chromosome", "GenomicDeletionRange"],
            ["GeneSymbol", "GenomicDeletionRange"],
            ["GenomicDeletionRange", "GeneSymbol"],
            ["HGVS", "GenomicDeletionRange"],
            ["ReferenceSequence", "GenomicDeletionRange"],
            ["LocusReferenceGenomic", "GenomicDeletionRange"]
        ]



"""Module for testing Protein Substitution Validator."""
import unittest

from variation.validators import ProteinSubstitution
from variation.classifiers import ProteinSubstitutionClassifier
from .validator_base import ValidatorBase


class TestProteinSubstitutionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Protein Substitution Validator."""

    def validator_instance(self):
        """Return protein substitution instance."""
        return ProteinSubstitution(*self.aa_params)

    def classifier_instance(self):
        """Return the protein substitution classifier instance."""
        return ProteinSubstitutionClassifier()

    def fixture_name(self) -> str:
        """Return the fixture name for protein substitution."""
        return "protein_substitution"



"""A module for Reference Sequence Tokenization."""
from typing import Optional

from variation.schemas.token_response_schema import Token, TokenMatchType
from .tokenizer import Tokenizer

REFSEQ_PREFIXES = ["NC_", "NT_", "NW_", "NG_", "NM_", "NR_", "NP_",
                   "ENSP", "ENST"]


class ReferenceSequence(Tokenizer):
    """Class for Reference Sequence."""

    def match(self, input_string: str) -> Optional[Token]:
        """Return a Reference Sequence match if it exists.

        :param str input_string: The input string to match
        """
        if input_string[:3].upper() in REFSEQ_PREFIXES or \
                input_string[:4] in REFSEQ_PREFIXES:
            return Token(
                token=input_string,
                token_type="ReferenceSequence",
                input_string=input_string,
                match_type=TokenMatchType.UNSPECIFIED
            )
        return None



"""A module for tokenizing."""
from typing import Iterable, List

from variation.schemas.token_response_schema import Token, TokenMatchType
from variation.tokenizers.caches.amino_acid_cache import AminoAcidCache
from .gene_symbol import GeneSymbol
from .protein_substitution import ProteinSubstitution
from .polypeptide_truncation import PolypeptideTruncation
from .silent_mutation import SilentMutation
from .coding_dna_substitution import CodingDNASubstitution
from .genomic_substitution import GenomicSubstitution
from .coding_dna_silent_mutation import CodingDNASilentMutation
from .genomic_silent_mutation import GenomicSilentMutation
from .protein_delins import ProteinDelIns
from .coding_dna_delins import CodingDNADelIns
from .genomic_delins import GenomicDelIns
from .hgvs import HGVS
from .reference_sequence import ReferenceSequence
from .locus_reference_genomic import LocusReferenceGenomic
from .protein_deletion import ProteinDeletion
from .coding_dna_deletion import CodingDNADeletion
from .genomic_deletion import GenomicDeletion
from .protein_insertion import ProteinInsertion
from .coding_dna_insertion import CodingDNAInsertion
from .genomic_insertion import GenomicInsertion
from .genomic_uncertain_deletion import GenomicUncertainDeletion
from .genomic_duplication import GenomicDuplication
from .genomic_deletion_range import GenomicDeletionRange
from .gnomad_vcf import GnomadVCF
from .caches import NucleotideCache


class Tokenize:
    """The tokenize class."""

    def __init__(self, amino_acid_cache: AminoAcidCache,
                 gene_symbol: GeneSymbol) -> None:
        """Initialize the tokenize class."""
        nucleotide_cache = NucleotideCache()
        self.tokenizers = (
            HGVS(),
            ReferenceSequence(),
            LocusReferenceGenomic(),
            GnomadVCF(),
            gene_symbol,
            ProteinSubstitution(amino_acid_cache),
            PolypeptideTruncation(amino_acid_cache),
            SilentMutation(amino_acid_cache),
            CodingDNASubstitution(),
            GenomicSubstitution(),
            CodingDNASilentMutation(),
            GenomicSilentMutation(),
            ProteinDelIns(amino_acid_cache, nucleotide_cache),
            CodingDNADelIns(amino_acid_cache, nucleotide_cache),
            GenomicDelIns(amino_acid_cache, nucleotide_cache),
            ProteinDeletion(amino_acid_cache, nucleotide_cache),
            CodingDNADeletion(amino_acid_cache, nucleotide_cache),
            GenomicDeletion(amino_acid_cache, nucleotide_cache),
            ProteinInsertion(amino_acid_cache, nucleotide_cache),
            CodingDNAInsertion(amino_acid_cache, nucleotide_cache),
            GenomicInsertion(amino_acid_cache, nucleotide_cache),
            GenomicUncertainDeletion(),
            GenomicDuplication(),
            GenomicDeletionRange()
        )

    def perform(self, search_string: str, warnings: List[str]) -> Iterable[Token]:
        """Return an iterable of tokens for a given search string.

        :param str search_string: The input string to search on
        :param List warnings: List of warnings
        :return: An Iterable of Tokens
        """
        tokens: List[Token] = list()
        # Currently splits on whitespace
        # Also adds a token if an string looks like an accession
        #  ex: NC_, NM_, ENST
        terms = search_string.split()
        self._add_tokens(tokens, terms, search_string, warnings)
        return tokens

    def _add_tokens(self, tokens: List, terms: List, search_string: str,
                    warnings: List) -> None:
        """Add tokens to a list for a given search string.

        :param List tokens: A list of tokens
        :param List terms: List of terms from query
        :param str search_string: The input string to search on
        :param List warnings: List of warnings
        """
        for term in terms:
            if not term:
                continue
            matched = False
            for tokenizer in self.tokenizers:
                res = tokenizer.match(term)
                if res:
                    if isinstance(res, List):
                        for r in res:
                            tokens.append(r)
                            if not matched:
                                matched = True
                    else:
                        tokens.append(res)
                        token = list(map(lambda t: t.token_type, tokens))[0]
                        if token == "HGVS" or \
                                token == "LocusReferenceGenomic" \
                                or token == "ReferenceSequence":
                            # Give specific type of HGVS (i.e. protein sub)
                            if len(tokens) == 1:
                                self._add_tokens(tokens,
                                                 [search_string.split(":")[1]],
                                                 search_string, warnings)
                        matched = True
                        break
                else:
                    continue
            if not matched:
                warnings.append(f"Unable to tokenize {term}")
                tokens.append(Token(
                    token="",
                    token_type="Unknown",
                    input_string=term,
                    match_type=TokenMatchType.UNSPECIFIED
                ))



"""A module for Genomic DelIns Tokenization."""
from typing import Dict, Optional

from variation.schemas.token_response_schema import GenomicDelInsToken
from variation.tokenizers.delins_base import DelInsBase


class GenomicDelIns(DelInsBase):
    """Class for tokenizing DelIns at the linear
    genomic reference sequence.
    """

    def return_token(self, params: Dict) -> Optional[GenomicDelInsToken]:
        """Return genomic delins token."""
        if self.parts["coordinate_type"] == "g":
            return GenomicDelInsToken(**params)



"""A module for testing Polypeptide Truncation tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache
from variation.tokenizers import PolypeptideTruncation
from .tokenizer_base import TokenizerBase


class TestPolypeptideTruncationTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Polypeptide Truncation  Tokenization."""

    def tokenizer_instance(self):
        """Return Polypeptide Truncation instance."""
        return PolypeptideTruncation(AminoAcidCache())

    def token_type(self):
        """Return Polypeptide Truncation token type."""
        return "PolypeptideTruncation"

    def fixture_name(self):
        """Return the fixture name for Polypeptide Truncation."""
        return "polypeptide_truncation"



"""A module for testing Genomic Substitution tokenization."""
import unittest

from variation.tokenizers import GenomicSubstitution
from .tokenizer_base import TokenizerBase


class TestGenomicSubstitutionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Genomic Substitution Tokenization."""

    def tokenizer_instance(self):
        """Return Genomic Substitution instance."""
        return GenomicSubstitution()

    def token_type(self):
        """Return genomic substitution token type."""
        return "GenomicSubstitution"

    def fixture_name(self):
        """Return the fixture name for genomic substitution."""
        return "genomic_substitution"



"""Module for testing protein Substitution Translator."""
import unittest

from variation.classifiers import ProteinSubstitutionClassifier
from variation.translators import ProteinSubstitution
from variation.validators import ProteinSubstitution as AASUB_V
from .translator_base import TranslatorBase


class TestProteinSubstitutionTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the protein Substitution Translator."""

    def classifier_instance(self):
        """Return protein substitution instance."""
        return ProteinSubstitutionClassifier()

    def validator_instance(self):
        """Return protein substitution instance."""
        return AASUB_V(*self.aa_params)

    def translator_instance(self):
        """Return protein substitution instance."""
        return ProteinSubstitution()

    def fixture_name(self):
        """Return the fixture name for protein substitution."""
        return "protein_substitution"



"""A module for testing Genomic Deletion Tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache, NucleotideCache
from variation.tokenizers import GenomicDeletion
from .tokenizer_base import TokenizerBase


class TestGenomicDeletionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Coding DNA Deletion Tokenization."""

    def tokenizer_instance(self):
        """Return Genomic Deletion instance."""
        return GenomicDeletion(AminoAcidCache(), NucleotideCache())

    def token_type(self):
        """Return genomic deletion token type."""
        return "GenomicDeletion"

    def fixture_name(self):
        """Return the fixture name for Genomic Deletion."""
        return "genomic_deletion"



"""Module for testing protein delins Translator."""
import unittest

from variation.classifiers import ProteinDelInsClassifier
from variation.translators import ProteinDelIns
from variation.validators import ProteinDelIns as AAD_V
from .translator_base import TranslatorBase


class TestProteinDelInsTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the protein delins Translator."""

    def classifier_instance(self):
        """Return protein delins instance."""
        return ProteinDelInsClassifier()

    def validator_instance(self):
        """Return protein delins instance."""
        return AAD_V(*self.aa_params)

    def translator_instance(self):
        """Return protein delins instance."""
        return ProteinDelIns()

    def fixture_name(self):
        """Return the fixture name for protein delins."""
        return "protein_delins"



"""Module for testing Genomic DelIns Validator."""
import unittest

from variation.validators import GenomicDelIns
from variation.classifiers import GenomicDelInsClassifier
from .validator_base import ValidatorBase


class TestGenomicDelInsValidator(ValidatorBase, unittest.TestCase):
    """A class to test the GenomicDelIns Validator."""

    def validator_instance(self):
        """Return genomic delins instance."""
        return GenomicDelIns(*self.params)

    def classifier_instance(self):
        """Return the genomic delins classifier instance."""
        return GenomicDelInsClassifier()

    def fixture_name(self):
        """Return the fixture name for genomic delins."""
        return "genomic_delins"



"""Module for testing Genomic DelIns Classifier."""
import unittest

from variation.classifiers import GenomicDelInsClassifier
from .classifier_base import ClassifierBase


class TestGenomicDelInsClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Genomic DelIns Classifier."""

    def classifier_instance(self):
        """Return GenomicDelInsClassifier instance."""
        return GenomicDelInsClassifier()

    def fixture_name(self):
        """Return GenomicDelInsClassifier fixture name."""
        return "genomic_delins"



"""Module for testing protein deletion Translator."""
import unittest

from variation.classifiers import ProteinDeletionClassifier
from variation.translators import ProteinDeletion
from variation.validators import ProteinDeletion as AAD_V
from .translator_base import TranslatorBase


class TestProteinDeletionTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the protein deletion Translator."""

    def classifier_instance(self):
        """Return protein deletion instance."""
        return ProteinDeletionClassifier()

    def validator_instance(self):
        """Return protein deletion instance."""
        return AAD_V(*self.aa_params)

    def translator_instance(self):
        """Return protein deletion instance."""
        return ProteinDeletion()

    def fixture_name(self):
        """Return the fixture name for protein deletion."""
        return "protein_deletion"



"""The module for Genomic Deletion Range Validation."""
import logging
from typing import List, Optional, Dict, Tuple

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass
from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from ga4gh.vrs.extras.translator import Translator
from ga4gh.vrs import models
from gene.query import QueryHandler as GeneQueryHandler
from uta_tools.data_sources import SeqRepoAccess, TranscriptMappings, UTADatabase, \
    MANETranscript

from variation.schemas.classification_response_schema import \
    ClassificationType, Classification
from variation.schemas.app_schemas import Endpoint
from variation.schemas.token_response_schema import Token
from variation.schemas.token_response_schema import GeneMatchToken
from variation.hgvs_dup_del_mode import HGVSDupDelMode
from variation.tokenizers import GeneSymbol
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum
from variation.validators.duplication_deletion_base import\
    DuplicationDeletionBase
from variation.vrs import VRS

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class GenomicDeletionRange(DuplicationDeletionBase):
    """The Genomic Deletion Range Validator class."""

    def __init__(self, seq_repo_access: SeqRepoAccess,
                 transcript_mappings: TranscriptMappings,
                 gene_symbol: GeneSymbol,
                 mane_transcript: MANETranscript,
                 uta: UTADatabase, dp: SeqRepoDataProxy, tlr: Translator,
                 gene_normalizer: GeneQueryHandler, vrs: VRS) -> None:
        """Initialize the Genomic Deletion Range validator.

        :param SeqRepoAccess seq_repo_access: Access to SeqRepo data
        :param TranscriptMappings transcript_mappings: Access to transcript
            mappings
        :param GeneSymbol gene_symbol: Gene symbol tokenizer
        :param MANETranscript mane_transcript: Access MANE Transcript
            information
        :param UTADatabase uta: Access to UTA queries
        :param GeneQueryHandler gene_normalizer: Access to gene-normalizer
        :param VRS vrs: Class for creating VRS objects
        """
        super().__init__(
            seq_repo_access, transcript_mappings, gene_symbol, mane_transcript,
            uta, dp, tlr, gene_normalizer, vrs
        )
        self.hgvs_dup_del_mode = HGVSDupDelMode(seq_repo_access)

    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        transcripts = await self.get_genomic_transcripts(classification, errors)
        return transcripts

    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        valid_alleles = list()
        for s in classification_tokens:
            for t in transcripts:
                errors = list()
                t = self.get_accession(t, classification)

                variation = await self._get_variation(
                    s, t, errors, gene_tokens, hgvs_dup_del_mode,
                    relative_copy_class=relative_copy_class,
                    baseline_copies=baseline_copies)

                if not errors and (endpoint_name == Endpoint.NORMALIZE or do_liftover):
                    await self._get_normalize_variation(
                        gene_tokens, s, t, errors, hgvs_dup_del_mode,
                        mane_data_found, relative_copy_class=relative_copy_class,
                        baseline_copies=baseline_copies)

                self.add_validation_result(
                    variation, valid_alleles, results,
                    classification, s, t, gene_tokens, errors
                )

                if is_identifier:
                    break

        if endpoint_name == Endpoint.NORMALIZE or do_liftover:
            self.add_mane_to_validation_results(
                mane_data_found, valid_alleles, results,
                classification, gene_tokens
            )

    async def _get_variation(
            self, s: Token, t: str, errors: List, gene_tokens: List,
            hgvs_dup_del_mode: HGVSDupDelModeEnum,
            relative_copy_class: Optional[RelativeCopyClass] = None,
            baseline_copies: Optional[int] = None) -> Optional[Dict]:
        """Get variation data.

        :param Token s: Classification token
        :param str t: Accession
        :param List errors: List of errors
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param HGVSDupDelMode hgvs_dup_del_mode: Mode to use for interpreting
            HGVS duplications and deletions
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param Optional[int] baseline_copies: Baseline copies number
        :return: Dictionary containing start/end position changes and variation
        """
        variation, start, end = None, None, None
        ival, grch38 = await self._get_ival(t, s, errors, gene_tokens)

        if not errors:
            if grch38:
                t = grch38["ac"]

            allele = self.vrs.to_vrs_allele_ranges(
                t, s.coordinate_type, s.alt_type, errors, ival)

            if start is not None and end is not None:
                pos = (start, end)
            else:
                pos = None

            variation = self.hgvs_dup_del_mode.interpret_variation(
                t, s.alt_type, allele, errors,
                hgvs_dup_del_mode, pos=pos, relative_copy_class=relative_copy_class,
                baseline_copies=baseline_copies)
        return variation

    async def _get_normalize_variation(
            self, gene_tokens: List, s: Token, t: str, errors: List,
            hgvs_dup_del_mode: HGVSDupDelModeEnum,
            mane_data_found: Dict,
            relative_copy_class: Optional[RelativeCopyClass] = None,
            baseline_copies: Optional[int] = None) -> None:
        """Get variation that will be returned in normalize endpoint.

        :param List gene_tokens: List of gene tokens
        :param Token s: Classification token
        :param str t: Accession
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Mode to use for
            interpreting HGVS duplications and deletions
        :param Dict mane_data_found: MANE Transcript data found for given query
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param Optional[int] baseline_copies: Baseline copies number
        """
        ival, grch38 = await self._get_ival(t, s, errors, gene_tokens, is_norm=True)
        self.add_grch38_to_mane_data(
            t, s, errors, grch38, mane_data_found, hgvs_dup_del_mode,
            ival=ival, relative_copy_class=relative_copy_class,
            baseline_copies=baseline_copies)

    async def _get_ival(
            self, t: str, s: Token, errors: List, gene_tokens: List,
            is_norm: bool = False
    ) -> Optional[Tuple[Optional[models.SequenceInterval], Optional[Dict]]]:
        """Get ival for variations with ranges.

        :param str t: Accession
        :param Token t: Classification token
        :param List errors: List of errors
        :param List gene_tokens: LIst of gene tokens
        :param bool is_norm: `True` if normalize endpoint is being used.
            `False` otherwise.
        :return: Sequence Interval and GRCh38 data if normalize endpoint
            is being used
        """
        ival, grch38, start1, start2, end1, end2 = None, None, None, None, None, None  # noqa: E501
        # (#_#)_(#_#)
        if is_norm:
            t, start1, start2, end1, end2, grch38 = await self.get_grch38_pos_ac(
                t, s.start_pos1_del, s.start_pos2_del, pos3=s.end_pos1_del,
                pos4=s.end_pos2_del
            )
        else:
            start1 = s.start_pos1_del
            start2 = s.start_pos2_del
            end1 = s.end_pos1_del
            end2 = s.end_pos2_del

        gene = gene_tokens[0].token if gene_tokens else None
        await self.validate_gene_or_accession_pos(
            t, [start1, start2, end1, end2], errors, gene=gene)

        if not errors and start1 and start2 and end1 and end2:
            ival = self.vrs.get_ival_certain_range(
                start1, start2, end1, end2
            )

        return ival, grch38

    def get_gene_tokens(
            self, classification: Classification) -> List[GeneMatchToken]:
        """Return gene tokens for a classification.

        :param Classification classification: The classification for tokens
        :return: A list of Gene Match Tokens in the classification
        """
        return self.get_gene_symbol_tokens(classification)

    def variation_name(self) -> str:
        """Return the variation name."""
        return "genomic deletion range"

    def is_token_instance(self, t: Token) -> bool:
        """Check that token is Genomic Deletion Range.

        :param Token t: Classification token
        """
        return t.token_type == "GenomicDeletionRange"

    def validates_classification_type(
            self,
            classification_type: ClassificationType) -> bool:
        """Return whether or not the classification type is
        Genomic Deletion Range.

        :param ClassificationType classification_type: Classification type
        :return: `True` if classification type matches, `False` otherwise
        """
        return classification_type == \
            ClassificationType.GENOMIC_DELETION_RANGE



"""Classifier package level import."""
from .classifier import Classifier
from .set_based_classifier import SetBasedClassifier
from .protein_delins_classifier import ProteinDelInsClassifier
from .protein_termination_classifier import ProteinTerminationClassifier
from .protein_classifier import ProteinSubstitutionClassifier
from .polypeptide_truncation_classifier import PolypeptideTruncationClassifier
from .silent_mutation import SilentMutationClassifier
from .coding_dna_substitution_classifier import CodingDNASubstitutionClassifier
from .genomic_substitution_classifier import GenomicSubstitutionClassifier
from .coding_dna_silent_mutation_classifier import CodingDNASilentMutationClassifier
from .genomic_silent_mutation_classifier import GenomicSilentMutationClassifier
from .coding_dna_delins_classifier import CodingDNADelInsClassifier
from .genomic_delins_classifier import GenomicDelInsClassifier
from .protein_deletion_classifier import ProteinDeletionClassifier
from .coding_dna_deletion_classifier import CodingDNADeletionClassifier
from .genomic_deletion_classifier import GenomicDeletionClassifier
from .protein_insertion_classifier import ProteinInsertionClassifier
from .coding_dna_insertion_classifier import CodingDNAInsertionClassifier
from .genomic_insertion_classifier import GenomicInsertionClassifier
from .genomic_uncertain_deletion_classifier import GenomicUncertainDeletionClassifier
from .genomic_duplication_classifier import GenomicDuplicationClassifier
from .genomic_deletion_range_classifier import GenomicDeletionRangeClassifier
from .classify import Classify



"""The module for Genomic Duplication Validation."""
import logging
from typing import List, Optional, Dict, Tuple

from ga4gh.vrs import models
from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass

from variation.schemas.app_schemas import Endpoint
from variation.validators.duplication_deletion_base import\
    DuplicationDeletionBase
from variation.schemas.classification_response_schema import \
    ClassificationType, Classification
from variation.schemas.token_response_schema import \
    TokenType, DuplicationAltType, Token, SequenceOntology
from variation.schemas.token_response_schema import GeneMatchToken
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class GenomicDuplication(DuplicationDeletionBase):
    """The Genomic Duplication Validator class."""

    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        transcripts = await self.get_genomic_transcripts(classification, errors)
        return transcripts

    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        valid_alleles = list()
        for s in classification_tokens:
            for t in transcripts:
                errors = list()
                t = self.get_accession(t, classification)

                result = await self._get_variation(
                    s, t, errors, gene_tokens, hgvs_dup_del_mode,
                    gene=gene_tokens[0].token if gene_tokens else None,
                    baseline_copies=baseline_copies,
                    relative_copy_class=relative_copy_class)
                variation = result["variation"]
                start = result["start"]
                end = result["end"]

                if not errors and (endpoint_name == Endpoint.NORMALIZE or do_liftover):
                    await self._get_normalize_variation(
                        gene_tokens, s, t, errors, hgvs_dup_del_mode,
                        mane_data_found, start, end,
                        relative_copy_class=relative_copy_class,
                        baseline_copies=baseline_copies)

                self.add_validation_result(
                    variation, valid_alleles, results,
                    classification, s, t, gene_tokens, errors
                )

                if is_identifier:
                    break

        if endpoint_name == Endpoint.NORMALIZE or do_liftover:
            self.add_mane_to_validation_results(
                mane_data_found, valid_alleles, results,
                classification, gene_tokens
            )

    async def _get_variation(
        self, s: Token, t: str, errors: List, gene_tokens: List,
        hgvs_dup_del_mode: HGVSDupDelModeEnum, gene: str = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None
    ) -> Optional[Dict]:
        """Get variation data.

        :param Token s: Classification token
        :param str t: Accession
        :param List errors: List of errors
        :param HGVSDupDelMode hgvs_dup_del_mode: Mode to use for interpreting
            HGVS duplications and deletions
        :param str gene: Gene symbol token
        :return: Dictionary containing start/end position changes and variation
        """
        variation, start, end = None, None, None
        if s.token_type == TokenType.GENOMIC_DUPLICATION:
            start = s.start_pos1_dup
            if s.start_pos2_dup is None:
                # Format: #dup
                end = s.start_pos1_dup
            else:
                # Format: #_#dup
                end = s.start_pos2_dup

            await self.validate_gene_or_accession_pos(
                t, [start, end], errors, gene=gene)

            if not errors:
                allele = self.vrs.to_vrs_allele(
                    t, start, end, s.coordinate_type,
                    s.alt_type, errors)
                variation = self.hgvs_dup_del_mode.interpret_variation(
                    t, s.alt_type, allele, errors, hgvs_dup_del_mode,
                    pos=(start, end), baseline_copies=baseline_copies,
                    relative_copy_class=relative_copy_class)
        elif s.token_type == TokenType.GENOMIC_DUPLICATION_RANGE:
            ival, grch38 = await self._get_ival(t, s, gene_tokens, errors)

            if not errors:
                if grch38:
                    t = grch38["ac"]

                allele = self.vrs.to_vrs_allele_ranges(
                    t, s.coordinate_type, s.alt_type, errors, ival)
                if start is not None and end is not None:
                    pos = (start, end)
                else:
                    pos = None
                variation = self.hgvs_dup_del_mode.interpret_variation(
                    t, s.alt_type, allele, errors,
                    hgvs_dup_del_mode, pos=pos, baseline_copies=baseline_copies,
                    relative_copy_class=relative_copy_class)
        else:
            errors.append(f"Token type not supported: {s.token_type}")

        return {
            "start": start,
            "end": end,
            "variation": variation
        }

    async def _get_normalize_variation(
        self, gene_tokens: List, s: Token, t: str, errors: List,
        hgvs_dup_del_mode: HGVSDupDelModeEnum, mane_data_found: Dict, start: int,
        end: int, baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None
    ) -> None:
        """Get variation that will be returned in normalize endpoint.

        :param List gene_tokens: List of gene tokens
        :param Token s: Classification token
        :param str t: Accession
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Mode to use for
            interpreting HGVS duplications and deletions
        :param Dict mane_data_found: MANE Transcript data found for given query
        :param int start: Start pos change
        :param int end: End pos change
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param Optional[int] baseline_copies: Baseline copies number
        """
        if s.token_type == TokenType.GENOMIC_DUPLICATION_RANGE:
            # (#_#)_(#_#)
            ival, grch38 = await self._get_ival(t, s, gene_tokens, errors, is_norm=True)
            self.add_grch38_to_mane_data(
                t, s, errors, grch38, mane_data_found, hgvs_dup_del_mode, ival=ival,
                baseline_copies=baseline_copies,
                relative_copy_class=relative_copy_class)
        else:
            # #dup or #_#dup
            if gene_tokens:
                gene = gene_tokens[0].token

                # Validate position
                await self._validate_gene_pos(gene, t, start, end, errors)
                if errors:
                    return

                await self.add_normalized_genomic_dup_del(
                    s, t, start, end, gene_tokens[0].token,
                    SequenceOntology.DUPLICATION, errors, hgvs_dup_del_mode,
                    mane_data_found, baseline_copies=baseline_copies,
                    relative_copy_class=relative_copy_class)
            else:
                grch38 = await self.mane_transcript.g_to_grch38(
                    t, start, end)

                if grch38:
                    await self.validate_gene_or_accession_pos(
                        grch38["ac"], [grch38["pos"][0], grch38["pos"][1]],
                        errors)
                    self.add_grch38_to_mane_data(
                        t, s, errors, grch38, mane_data_found,
                        hgvs_dup_del_mode, use_vrs_allele_range=False,
                        baseline_copies=baseline_copies,
                        relative_copy_class=relative_copy_class
                    )

    async def _get_ival(
            self, t: str, s: Token, gene_tokens: List, errors: List,
            is_norm: bool = False
    ) -> Optional[Tuple[models.SequenceInterval, Dict]]:
        """Get ival for variations with ranges.

        :param str t: Accession
        :param Token t: Classification token
        :param List gene_tokens: List of gene tokens
        :param List errors: List of errors
        :param bool is_norm: `True` if normalize endpoint is being used.
            `False` otherwise.
        :return: Sequence Interval and GRCh38 data if normalize endpoint
            is being used
        """
        ival, start, end, grch38 = None, None, None, None
        gene = gene_tokens[0].token if gene_tokens else None
        if s.alt_type != DuplicationAltType.UNCERTAIN_DUPLICATION:
            # (#_#)_(#_#)
            if is_norm:
                t, start1, start2, end1, end2, grch38 = await self.get_grch38_pos_ac(
                    t, s.start_pos1_dup, s.start_pos2_dup, pos3=s.end_pos1_dup,
                    pos4=s.end_pos2_dup
                )
            else:
                start1 = s.start_pos1_dup
                start2 = s.start_pos2_dup
                end1 = s.end_pos1_dup
                end2 = s.end_pos2_dup

            if start1 and start2 and end1 and end2:
                await self.validate_gene_or_accession_pos(
                    t, [start1, start2, end1, end2], errors, gene=gene)

                if not errors:
                    ival = self.vrs.get_ival_certain_range(
                        start1, start2, end1, end2)
        else:
            if s.start_pos1_dup == "?" and s.end_pos2_dup == "?":
                # format: (?_#)_(#_?)
                if is_norm:
                    t, start, end, _, _, grch38 = await self.get_grch38_pos_ac(
                        t, s.start_pos2_dup, s.end_pos1_dup
                    )
                else:
                    start = s.start_pos2_dup
                    end = s.end_pos1_dup

                # Validate positions
                await self.validate_gene_or_accession_pos(
                    t, [start, end], errors, gene=gene)

                if not errors and start and end:
                    ival = models.SequenceInterval(
                        start=self.vrs.get_start_indef_range(start),
                        end=self.vrs.get_end_indef_range(end),
                        type="SequenceInterval"
                    )
            elif s.start_pos1_dup == "?" and \
                    s.start_pos2_dup != "?" and \
                    s.end_pos1_dup != "?" and \
                    s.end_pos2_dup is None:
                # format: (?_#)_#
                if is_norm:
                    t, start, end, _, _, grch38 = await self.get_grch38_pos_ac(
                        t, s.start_pos2_dup, s.end_pos1_dup)
                else:
                    start = s.start_pos2_dup
                    end = s.end_pos1_dup

                # Validate positions
                await self.validate_gene_or_accession_pos(
                    t, [start, end], errors, gene=gene)

                if not errors and start and end:
                    ival = models.SequenceInterval(
                        start=self.vrs.get_start_indef_range(start),
                        end=models.Number(value=end, type="Number"),
                        type="SequenceInterval"
                    )
            elif s.start_pos1_dup != "?" and \
                    s.start_pos2_dup is None and \
                    s.end_pos1_dup != "?" and \
                    s.end_pos2_dup == "?":
                # format: #_(#_?)
                if is_norm:
                    t, start, end, _, _, grch38 = await self.get_grch38_pos_ac(
                        t, s.start_pos1_dup, s.end_pos1_dup
                    )
                else:
                    start = s.start_pos1_dup
                    end = s.end_pos1_dup
                start -= 1

                await self.validate_gene_or_accession_pos(
                    t, [start, end], errors, gene=gene)

                if not errors and start and end:
                    ival = models.SequenceInterval(
                        start=models.Number(value=start, type="Number"),
                        end=self.vrs.get_end_indef_range(end),
                        type="SequenceInterval"
                    )
            else:
                errors.append("Not yet supported")
        return ival, grch38

    def get_gene_tokens(
            self, classification: Classification) -> List[GeneMatchToken]:
        """Return gene tokens for a classification.

        :param Classification classification: The classification for tokens
        :return: A list of Gene Match Tokens in the classification
        """
        return self.get_gene_symbol_tokens(classification)

    def variation_name(self) -> str:
        """Return the variation name."""
        return "genomic duplication"

    def is_token_instance(self, t: Token) -> bool:
        """Check that token is an instance of Genomic Duplication."""
        return t.token_type in ["GenomicDuplication",
                                "GenomicDuplicationRange"]

    def validates_classification_type(
        self, classification_type: ClassificationType
    ) -> bool:
        """Return whether or not the classification type is
        Genomic Duplication.

        :param ClassificationType classification_type: Classification type
        :return: `True` if classification type matches, `False` otherwise
        """
        return classification_type == \
            ClassificationType.GENOMIC_DUPLICATION



"""Module for Protein Alternate classification."""
import re
from typing import Optional

from variation.schemas.token_response_schema import Token, TokenMatchType
from .tokenizer import Tokenizer
from .caches import AminoAcidCache


class ProteinAlternate(Tokenizer):
    """The Protein Alternate Tokenization class."""

    def __init__(self, amino_acid_cache: AminoAcidCache) -> None:
        """Initialize the Protein Alternate Tokenizer class."""
        self.__amino_acid_cache = amino_acid_cache
        self.__splitter = re.compile(r"\d+")

    def match(self, input_string: str) -> Optional[Token]:
        """Return Protein Alternate tokens if input string matches."""
        potential_protein = self.__splitter.split(input_string)
        if (len(potential_protein) == 2 and  # noqa: W504
                potential_protein[0] in self.__amino_acid_cache and   # noqa: W504
                not potential_protein[1]):
            return Token(
                token=potential_protein[0],
                token_type="ProteinAlternate",
                input_string=input_string,
                match_type=TokenMatchType.UNSPECIFIED
            )
        else:
            return None



"""Module for classification."""
from typing import List, Callable

from variation.schemas.classification_response_schema import Classification, \
    ConfidenceRating
from variation.schemas.token_response_schema import Token
from variation.classifiers import ProteinTerminationClassifier, \
    ProteinSubstitutionClassifier, PolypeptideTruncationClassifier, \
    SilentMutationClassifier, CodingDNASubstitutionClassifier,\
    GenomicSubstitutionClassifier, CodingDNASilentMutationClassifier, \
    GenomicSilentMutationClassifier, ProteinDelInsClassifier, \
    CodingDNADelInsClassifier, GenomicDelInsClassifier, \
    ProteinDeletionClassifier, CodingDNADeletionClassifier, \
    GenomicDeletionClassifier, ProteinInsertionClassifier, \
    CodingDNAInsertionClassifier, GenomicInsertionClassifier, \
    GenomicUncertainDeletionClassifier, GenomicDuplicationClassifier, \
    GenomicDeletionRangeClassifier, Classifier


class Classify:
    """The classify class."""

    def __init__(self) -> None:
        """Initialize the Classify class."""
        self.classifiers: List[Classifier] = [
            ProteinDelInsClassifier(),
            ProteinSubstitutionClassifier(),
            PolypeptideTruncationClassifier(),
            SilentMutationClassifier(),
            ProteinTerminationClassifier(),
            CodingDNASubstitutionClassifier(),
            GenomicSubstitutionClassifier(),
            CodingDNASilentMutationClassifier(),
            GenomicSilentMutationClassifier(),
            ProteinDelInsClassifier(),
            CodingDNADelInsClassifier(),
            GenomicDelInsClassifier(),
            ProteinDeletionClassifier(),
            CodingDNADeletionClassifier(),
            GenomicDeletionClassifier(),
            ProteinInsertionClassifier(),
            CodingDNAInsertionClassifier(),
            GenomicInsertionClassifier(),
            GenomicUncertainDeletionClassifier(),
            GenomicDuplicationClassifier(),
            GenomicDeletionRangeClassifier()
        ]

    def perform(self, tokens: List[Token]) -> List[Classification]:
        """Classify a list of tokens.

        :param List[Token] tokens: List of tokens found
        :return: A classification with the highest confidence rating
        """
        classifications: List[Classification] = list()

        for classifier in self.classifiers:
            res = classifier.match(tokens)
            if res is not None and res.confidence == ConfidenceRating.EXACT:
                return [res]
            elif res is not None:
                classifications.append(res)

        highest_confidence = 0
        for match in classifications:
            if match.confidence.value > highest_confidence:
                highest_confidence = match.confidence.value

        filter_lambda: Callable[[Classification], bool] = \
            lambda match: match.confidence.value == highest_confidence

        return list(filter(filter_lambda, classifications))



"""A module for DelIns Tokenization Base Class."""
from abc import abstractmethod
from typing import Optional, Dict, List

from variation.schemas.token_response_schema import DelIns, TokenMatchType, Token
from .tokenizer import Tokenizer
from .caches import AminoAcidCache, NucleotideCache
from .tokenize_base import TokenizeBase


class DelInsBase(Tokenizer):
    """Class for tokenizing DelIns."""

    def __init__(self, amino_acid_cache: AminoAcidCache,
                 nucleotide_cache: NucleotideCache) -> None:
        """Initialize the DelIns Class.

        :param AminoAcidCache amino_acid_cache: Valid amino acid codes
        :param NucleotideCache nucleotide_cache: Valid nucleotides
        """
        self.parts = None
        self.tokenize_base = TokenizeBase(amino_acid_cache, nucleotide_cache)

    def match(self, input_string: str) -> Optional[DelIns]:
        """Return tokens that match the input string."""
        if input_string is None:
            return None

        input_string = str(input_string).lower()

        if input_string.startswith("(") and input_string.endswith(")"):
            input_string = input_string[1:-1]

        if ("delins" not in input_string) or (
                (not input_string.startswith("c.") and (not
                 input_string.startswith("g.")))):
            return None

        self.parts = {
            "start_pos_del": None,
            "end_pos_del": None,
            "inserted_sequence1": None,
            "inserted_sequence2": None,
            "coordinate_type": None
        }

        parts = input_string.split("delins")
        self._get_parts(parts)

        # TODO: implement delins range
        #  Ex: 812_829delins908_925
        if self.parts["inserted_sequence1"] is not None and \
                self.parts["inserted_sequence2"] is not None:
            return None
        params = {
            "token": input_string,
            "input_string": input_string,
            "match_type": TokenMatchType.UNSPECIFIED.value,
            "start_pos_del": self.parts["start_pos_del"],
            "end_pos_del": self.parts["end_pos_del"],
            "inserted_sequence1": self.parts["inserted_sequence1"],
            "inserted_sequence2": self.parts["inserted_sequence2"],
            "coordinate_type": self.parts["coordinate_type"]
        }
        return self.return_token(params)

    def _get_parts(self, parts: List) -> None:
        """Get parts for DelIns.

        :param List parts: Parts of input string
        """
        if len(parts) != 2:
            return

        if not parts[0].startswith("c.") and not parts[0].startswith("g."):
            return

        # Get reference sequence
        coordinate_type = parts[0][:1]
        parts[0] = parts[0][2:]

        positions_deleted = self.tokenize_base.get_positions_deleted(parts)
        if not positions_deleted:
            return
        start_pos_del = positions_deleted[0]
        end_pos_del = positions_deleted[1]

        inserted_sequences = \
            self.tokenize_base.get_transcript_genomic_inserted_sequence(parts)
        if not inserted_sequences:
            return
        inserted_sequence1 = inserted_sequences[0]
        inserted_sequence2 = inserted_sequences[1]

        if not inserted_sequence1 and not inserted_sequence2:
            return

        self.parts = {
            "start_pos_del": start_pos_del,
            "end_pos_del": end_pos_del,
            "inserted_sequence1": inserted_sequence1,
            "inserted_sequence2": inserted_sequence2,
            "coordinate_type": coordinate_type
        }

    @abstractmethod
    def return_token(self, params: Dict[str, str]) -> Optional[Token]:
        """Return token instance."""
        raise NotImplementedError



"""Module for testing Protein Insertion Classifier."""
import unittest

from variation.classifiers import ProteinInsertionClassifier
from .classifier_base import ClassifierBase


class TestProteinInsertionClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Protein Insertion Classifier."""

    def classifier_instance(self):
        """Return ProteinInsertionClassifier instance."""
        return ProteinInsertionClassifier()

    def fixture_name(self):
        """Return ProteinInsertionClassifier fixture name."""
        return "protein_insertion"



"""The module for Coding DNA Insertion Validation."""
from typing import List, Optional
import logging

from variation.validators.insertion_base import InsertionBase
from variation.schemas.classification_response_schema import \
    Classification, ClassificationType
from variation.schemas.token_response_schema import Token
from variation.schemas.token_response_schema import GeneMatchToken

logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class CodingDNAInsertion(InsertionBase):
    """The Coding DNA Insertion Validator class."""

    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        return self.get_coding_dna_transcripts(gene_tokens, errors)

    def get_gene_tokens(self, classification: Classification) -> List[GeneMatchToken]:
        """Return gene tokens for a classification.

        :param Classification classification: The classification for tokens
        :return: A list of Gene Match Tokens in the classification
        """
        return self.get_coding_dna_gene_symbol_tokens(classification)

    def variation_name(self) -> str:
        """Return the variation name."""
        return "coding dna insertion"

    def is_token_instance(self, t: Token) -> bool:
        """Check that token is Coding DNA Insertion."""
        return t.token_type == "CodingDNAInsertion"

    def validates_classification_type(
        self, classification_type: ClassificationType
    ) -> bool:
        """Return whether or not the classification type is
        Coding DNA Insertion.
        """
        return classification_type == ClassificationType.CODING_DNA_INSERTION



"""Module for testing Genomic Deletion Translator."""
import unittest

from variation.classifiers import GenomicDeletionClassifier
from variation.translators import GenomicDeletion
from variation.validators import GenomicDeletion as GD_V
from .translator_base import TranslatorBase


class TestGenomicDeletionTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Genomic Deletion Translator."""

    def classifier_instance(self):
        """Return genomic deletion instance."""
        return GenomicDeletionClassifier()

    def validator_instance(self):
        """Return coding DNA deletion instance."""
        return GD_V(*self.params)

    def translator_instance(self):
        """Return genomic deletion instance."""
        return GenomicDeletion()

    def fixture_name(self):
        """Return the fixture name for genomic deletion."""
        return "genomic_deletion"



"""A module for gnomad VCF tokenization"""
from typing import Optional, List, Dict
import re

from variation.schemas.token_response_schema import TokenMatchType, \
    Token, ChromosomeToken, GenomicSubstitutionToken, \
    GenomicSilentMutationToken, GenomicDeletionToken, GenomicInsertionToken, \
    Nomenclature
from .tokenizer import Tokenizer


class GnomadVCF(Tokenizer):
    """The gnomad VCF tokenizer class"""

    def __init__(self) -> None:
        """Initialize the gnomad VCF tokenizer class"""
        self.splitter = re.compile(
            r"^(?P<chromosome>(chr|chromosome)?([1-9]|[1][0-9]|[2][0-2]|X|Y))-"
            r"(?P<pos>[1-9]\d*)-(?P<ref>(?i)[actg]+)-(?P<alt>(?i)[actg]+)$")

    def match(self, input_string: str) -> Optional[List[Token]]:
        """Return a GnomadVCFToken if a match exists.

        :param str input_string: The input string to match
        :return: List of tokens
        """
        params = self._get_params(input_string)
        if not params:
            return None

        tokens = list()
        chromosome = params["chromosome"]
        if "chr" in chromosome:
            if "chromosome" in chromosome:
                chromosome_val = chromosome[10:]
            else:
                chromosome_val = chromosome[3:]
        else:
            chromosome_val = chromosome
        params["chromosome"] = f"chr{chromosome_val.upper()}"
        for field in ["ref", "alt"]:
            params[field] = params[field].upper()
        params["pos"] = int(params["pos"])

        tokens.append(ChromosomeToken(
            token=chromosome,
            input_string=input_string,
            match_type=TokenMatchType.UNSPECIFIED.value,
            chromosome=params["chromosome"],
            nomenclature=Nomenclature.GNOMAD_VCF)
        )

        ref_len = len(params["ref"])
        alt_len = len(params["alt"])

        if ref_len == alt_len:
            if ref_len == 1:
                if params["ref"] == params["alt"]:
                    tokens.append(GenomicSilentMutationToken(
                        token=input_string,
                        input_string=input_string,
                        match_type=TokenMatchType.UNSPECIFIED.value,
                        position=params["pos"],
                        ref_nucleotide=params["ref"],
                        nomenclature=Nomenclature.GNOMAD_VCF
                    ))
                else:
                    tokens.append(GenomicSubstitutionToken(
                        token=input_string,
                        input_string=input_string,
                        match_type=TokenMatchType.UNSPECIFIED.value,
                        position=params["pos"],
                        ref_nucleotide=params["ref"],
                        new_nucleotide=params["alt"],
                        nomenclature=Nomenclature.GNOMAD_VCF
                    ))
        elif ref_len < alt_len:
            # insertion
            if ref_len == 1:
                if params["ref"][0] == params["alt"][0]:
                    tokens.append(GenomicInsertionToken(
                        token=input_string,
                        input_string=input_string,
                        match_type=TokenMatchType.UNSPECIFIED.value,
                        start_pos_flank=params["pos"],
                        end_pos_flank=params["pos"] + 1,
                        inserted_sequence=params["alt"][1:],
                        nomenclature=Nomenclature.GNOMAD_VCF
                    ))
        else:
            # deletion
            if alt_len == 1:
                if params["ref"][0] == params["alt"][0]:
                    del_seq = params["ref"][1:]
                    tokens.append(GenomicDeletionToken(
                        token=input_string,
                        input_string=input_string,
                        match_type=TokenMatchType.UNSPECIFIED.value,
                        start_pos_del=params["pos"] + 1,
                        end_pos_del=params["pos"] + len(del_seq),
                        deleted_sequence=del_seq,
                        nomenclature=Nomenclature.GNOMAD_VCF
                    ))
        return tokens if tokens else None

    def _get_params(self, input_string: str) -> Optional[Dict]:
        """Get chromosome, pos, ref, alt params from input string.

        :param str input_string: Input string
        :return: chromosome, pos, ref, alt data
        """
        if not input_string:
            return None
        input_string = input_string.strip()

        match = self.splitter.match(input_string)
        if not match:
            return None
        else:
            return match.groupdict()



"""Module for testing Coding DNA Silent Mutation Classifier."""
import unittest

from variation.classifiers import CodingDNASilentMutationClassifier
from .classifier_base import ClassifierBase


class TestCodingDNASilentMutationClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Coding DNA Silent Mutation Classifier."""

    def classifier_instance(self):
        """Return CodingDNASilentMutationClassifier instance."""
        return CodingDNASilentMutationClassifier()

    def fixture_name(self):
        """Return CodingDNASilentMutationClassifier fixture name."""
        return "coding_dna_silent_mutation"



"""The module for Coding DNA Substitution Validation."""
import logging
from typing import List, Optional, Dict

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass

from variation.schemas.classification_response_schema import \
    ClassificationType, Classification
from variation.schemas.token_response_schema import Token
from variation.schemas.token_response_schema import GeneMatchToken
from variation.schemas.app_schemas import Endpoint
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum
from .single_nucleotide_variation_base import SingleNucleotideVariationBase

# TODO:
#  LRG_ (LRG_199t1:c)


logger = logging.getLogger("variation")
logger.setLevel(logging.DEBUG)


class CodingDNASubstitution(SingleNucleotideVariationBase):
    """The Coding DNA Substitution Validator class."""

    async def get_transcripts(self, gene_tokens: List, classification: Classification,
                              errors: List) -> Optional[List[str]]:
        """Get transcript accessions for a given classification.

        :param List gene_tokens: A list of gene tokens
        :param Classification classification: A classification for a list of
            tokens
        :param List errors: List of errors
        :return: List of transcript accessions
        """
        return self.get_coding_dna_transcripts(gene_tokens, errors)

    async def get_valid_invalid_results(
        self, classification_tokens: List, transcripts: List,
        classification: Classification, results: List, gene_tokens: List,
        mane_data_found: Dict, is_identifier: bool,
        hgvs_dup_del_mode: HGVSDupDelModeEnum,
        endpoint_name: Optional[Endpoint] = None,
        baseline_copies: Optional[int] = None,
        relative_copy_class: Optional[RelativeCopyClass] = None,
        do_liftover: bool = False
    ) -> None:
        """Add validation result objects to a list of results.

        :param List classification_tokens: A list of classification Tokens
        :param List transcripts: A list of transcript accessions
        :param Classification classification: A classification for a list of
            tokens
        :param List results: Stores validation result objects
        :param List gene_tokens: List of GeneMatchTokens for a classification
        :param Dict mane_data_found: MANE Transcript information found
        :param bool is_identifier: `True` if identifier is given for exact
            location. `False` otherwise.
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        """
        valid_alleles = list()
        for s in classification_tokens:
            for t in transcripts:
                errors = list()

                t = self.get_accession(t, classification)
                cds_start_end = await self.uta.get_cds_start_end(t)
                if cds_start_end is not None:
                    cds_start = cds_start_end[0]

                    allele = self.vrs.to_vrs_allele(
                        t, s.position, s.position, s.coordinate_type,
                        s.alt_type, errors, cds_start=cds_start,
                        alt=s.new_nucleotide)
                else:
                    allele = None
                    cds_start = 0
                    errors.append(f"Unable to get CDS start for {t}")

                if not errors:
                    ref_nuc, _ = self.seqrepo_access.get_reference_sequence(
                        t, s.position + cds_start)
                    self.check_ref_nucleotide(ref_nuc, s.ref_nucleotide,
                                              s.position, t, errors)

                if not errors and endpoint_name == Endpoint.NORMALIZE:
                    mane = await self.mane_transcript.get_mane_transcript(
                        t, s.position, s.coordinate_type, end_pos=s.position,
                        ref=s.ref_nucleotide, try_longest_compatible=True
                    )

                    self.add_mane_data(mane, mane_data_found, s.coordinate_type,
                                       s.alt_type, s, alt=s.new_nucleotide)

                self.add_validation_result(allele, valid_alleles, results,
                                           classification, s, t, gene_tokens, errors)

                if is_identifier:
                    break

        if endpoint_name == Endpoint.NORMALIZE:
            self.add_mane_to_validation_results(mane_data_found, valid_alleles, results,
                                                classification, gene_tokens)

    def get_gene_tokens(self, classification: Classification) -> List[GeneMatchToken]:
        """Return gene tokens for a classification.

        :param Classification classification: The classification for tokens
        :return: A list of Gene Match Tokens in the classification
        """
        return self.get_coding_dna_gene_symbol_tokens(classification)

    def variation_name(self) -> str:
        """Return the variation name."""
        return "coding dna substitution"

    def is_token_instance(self, t: Token) -> bool:
        """Check that token is Coding DNA Substitution."""
        return t.token_type == "CodingDNASubstitution"

    def validates_classification_type(
            self, classification_type: ClassificationType) -> bool:
        """Return whether or not the classification type is coding dna
        substitution.
        """
        return classification_type == ClassificationType.CODING_DNA_SUBSTITUTION



"""Module for Protein Insertion Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import ProteinInsertionToken, Token


class ProteinInsertion(Translator):
    """The Protein Insertion Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Protein Insertion."""
        return type == ClassificationType.PROTEIN_INSERTION

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Protein Insertion token instance."""
        return isinstance(token, ProteinInsertionToken)



"""Module for Genomic DelIns Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import GenomicDelInsToken, Token


class GenomicDelIns(Translator):
    """The Genomic DelIns Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Genomic DelIns."""
        return type == ClassificationType.GENOMIC_DELINS

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Genomic DelIns token instance."""
        return isinstance(token, GenomicDelInsToken)



"""A module for Deletion Tokenization Base Class."""
from abc import abstractmethod
from typing import Optional, Dict, List

from variation.schemas.token_response_schema import Deletion, TokenMatchType
from .tokenizer import Tokenizer
from .caches import AminoAcidCache, NucleotideCache
from .tokenize_base import TokenizeBase


class DeletionBase(Tokenizer):
    """Class for tokenizing Deletions."""

    def __init__(self, amino_acid_cache: AminoAcidCache,
                 nucleotide_cache: NucleotideCache) -> None:
        """Initialize the Deletion Base Class.

        :param AminoAcidCache amino_acid_cache: Valid amino acid codes
        :param NucleotideCache nucleotide_cache: Valid nucleotides
        """
        self.parts = None
        self.tokenize_base = TokenizeBase(amino_acid_cache, nucleotide_cache)

    def match(self, input_string: str) -> Optional[Deletion]:
        """Return tokens that match the input string.

        :param str input_string: Input string
        :return: Deletion token if a match is found
        """
        if input_string is None:
            return None

        self.parts = {
            "token": input_string,
            "input_string": input_string,
            "match_type": TokenMatchType.UNSPECIFIED.value,
            "start_pos_del": None,
            "end_pos_del": None,
            "deleted_sequence": None,
            "coordinate_type": None
        }

        input_string = str(input_string).lower()
        conditions = (
            "del" in input_string,
            "ins" not in input_string and "delins" not in input_string,
            input_string.startswith("c.") or input_string.startswith("g.")
        )
        if not all(conditions):
            return None

        parts = input_string.split("del")
        self._get_parts(parts)
        return self.return_token(self.parts)

    def _get_parts(self, parts: List) -> None:
        """Get parts for DelIns by updating `self.parts`

        :param List parts: Parts of input string
        """
        if len(parts) != 2:
            return None

        # Get reference sequence
        coordinate_type = parts[0][:1]
        parts[0] = parts[0][2:]

        positions_deleted = self.tokenize_base.get_positions_deleted(parts)
        if not positions_deleted:
            return None

        if parts[1]:
            self.parts["deleted_sequence"] = \
                self.tokenize_base.get_sequence(parts[1])

        if positions_deleted[0]:
            start_pos_del = int(positions_deleted[0])
        else:
            start_pos_del = None
        if positions_deleted[1]:
            end_pos_del = int(positions_deleted[1])
        else:
            end_pos_del = None

        if start_pos_del and end_pos_del:
            if start_pos_del > end_pos_del:
                return None

        self.parts["start_pos_del"] = start_pos_del
        self.parts["end_pos_del"] = end_pos_del
        self.parts["coordinate_type"] = coordinate_type

    @abstractmethod
    def return_token(self, params: Dict[str, str]) -> Optional[Deletion]:
        """Return token instance.

        :param Dict params: Params for Deletion token
        :return: Deletion token
        """
        raise NotImplementedError



"""Module for testing Genomic Substitution Translator."""
import unittest

from variation.classifiers import GenomicSubstitutionClassifier
from variation.translators import GenomicSubstitution
from variation.validators import GenomicSubstitution as GSUB_V
from .translator_base import TranslatorBase


class TestGenomicSubstitutionTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Genomic Substitution Translator."""

    def classifier_instance(self):
        """Return genomic substitution instance."""
        return GenomicSubstitutionClassifier()

    def validator_instance(self):
        """Return genomic substitution instance."""
        return GSUB_V(*self.params)

    def translator_instance(self):
        """Return genomic substitution instance."""
        return GenomicSubstitution()

    def fixture_name(self):
        """Return the fixture name for genomic substitution."""
        return "genomic_substitution"



"""A module for testing Coding DNA Substitution tokenization."""
import unittest

from variation.tokenizers import CodingDNASubstitution
from .tokenizer_base import TokenizerBase


class TestCodingDNASubstitutionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Coding DNA Substitution Tokenization."""

    def tokenizer_instance(self):
        """Return Coding DNA Substitution instance."""
        return CodingDNASubstitution()

    def token_type(self):
        """Return coding DNA substitution token type."""
        return "CodingDNASubstitution"

    def fixture_name(self):
        """Return the fixture name for coding DNA substitution."""
        return "coding_dna_substitution"



"""Module for Validation."""
from typing import List, Optional

from ga4gh.vrsatile.pydantic.vrs_models import RelativeCopyClass
from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from ga4gh.vrs.extras.translator import Translator
from gene.query import QueryHandler as GeneQueryHandler
from uta_tools.data_sources import TranscriptMappings, SeqRepoAccess, UTADatabase, \
    MANETranscript

from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum
from variation.vrs import VRS
from variation.schemas.app_schemas import Endpoint
from variation.schemas.validation_response_schema import ValidationSummary
from variation.schemas.classification_response_schema import Classification
from variation.tokenizers import GeneSymbol
from variation.tokenizers.caches import AminoAcidCache
from .protein_substitution import ProteinSubstitution
from .polypeptide_truncation import PolypeptideTruncation
from .silent_mutation import SilentMutation
from .coding_dna_substitution import CodingDNASubstitution
from .coding_dna_silent_mutation import CodingDNASilentMutation
from .genomic_silent_mutation import GenomicSilentMutation
from .genomic_substitution import GenomicSubstitution
from .protein_delins import ProteinDelIns
from .coding_dna_delins import CodingDNADelIns
from .genomic_delins import GenomicDelIns
from .protein_deletion import ProteinDeletion
from .coding_dna_deletion import CodingDNADeletion
from .genomic_deletion import GenomicDeletion
from .protein_insertion import ProteinInsertion
from .coding_dna_insertion import CodingDNAInsertion
from .genomic_insertion import GenomicInsertion
from .genomic_uncertain_deletion import GenomicUncertainDeletion
from .genomic_duplication import GenomicDuplication
from .genomic_deletion_range import GenomicDeletionRange


class Validate:
    """The validation class."""

    def __init__(self, seqrepo_access: SeqRepoAccess,
                 transcript_mappings: TranscriptMappings,
                 gene_symbol: GeneSymbol,
                 mane_transcript: MANETranscript,
                 uta: UTADatabase, dp: SeqRepoDataProxy, tlr: Translator,
                 amino_acid_cache: AminoAcidCache,
                 gene_normalizer: GeneQueryHandler, vrs: VRS) -> None:
        """Initialize the validate class.

        :param SeqRepoAccess seqrepo_access: Access to SeqRepo data
        :param TranscriptMappings transcript_mappings: Access to transcript
            mappings
        :param GeneSymbol gene_symbol: Gene symbol tokenizer
        :param MANETranscript mane_transcript: Access MANE Transcript
            information
        :param UTADatabase uta: Access to UTA queries
        :param Translator tlr: Translator class
        :param GeneQueryHandler gene_normalizer: Access to gene-normalizer
        :param VRS vrs: Class for creating VRS objects
        :param amino_acid_cache: Amino Acid codes and conversions
        """
        params = [
            seqrepo_access, transcript_mappings, gene_symbol,
            mane_transcript, uta, dp, tlr, gene_normalizer, vrs
        ]
        protein_params = params[:]
        protein_params.append(amino_acid_cache)
        self.validators = [
            ProteinSubstitution(*protein_params),
            PolypeptideTruncation(*protein_params),
            SilentMutation(*protein_params),
            CodingDNASubstitution(*params),
            GenomicSubstitution(*params),
            CodingDNASilentMutation(*params),
            GenomicSilentMutation(*params),
            ProteinDelIns(*protein_params),
            CodingDNADelIns(*params),
            GenomicDelIns(*params),
            ProteinDeletion(*protein_params),
            CodingDNADeletion(*params),
            GenomicDeletion(*params),
            ProteinInsertion(*protein_params),
            CodingDNAInsertion(*params),
            GenomicInsertion(*params),
            GenomicDeletionRange(*params),
            GenomicUncertainDeletion(*params),
            GenomicDuplication(*params)
        ]

    async def perform(
            self, classifications: List[Classification],
            endpoint_name: Optional[Endpoint] = None, warnings: List = None,
            hgvs_dup_del_mode: HGVSDupDelModeEnum = HGVSDupDelModeEnum.DEFAULT,
            baseline_copies: Optional[int] = None,
            relative_copy_class: Optional[RelativeCopyClass] = None,
            do_liftover: bool = False
    ) -> ValidationSummary:
        """Validate a list of classifications.

        :param List classifications: List of classifications
        :param Optional[Endpoint] endpoint_name: Then name of the endpoint being used
        :param List warnings: List of warnings
        :param HGVSDupDelModeEnum hgvs_dup_del_mode: Must be: `default`, `cnv`,
            `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to represent HGVS dup/del expressions
            as VRS objects.
        :param Optional[int] baseline_copies: Baseline copies number
        :param Optional[RelativeCopyClass] relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        :return: ValidationSummary containing valid and invalid results
        """
        valid_possibilities = list()
        invalid_possibilities = list()
        if not warnings:
            warnings = list()

        found_classification = False
        for classification in classifications:
            for validator in self.validators:
                if validator.validates_classification_type(
                        classification.classification_type):
                    results = await validator.validate(
                        classification, hgvs_dup_del_mode=hgvs_dup_del_mode,
                        endpoint_name=endpoint_name, baseline_copies=baseline_copies,
                        relative_copy_class=relative_copy_class,
                        do_liftover=do_liftover)
                    for res in results:
                        if res.is_valid:
                            found_classification = True
                            valid_possibilities.append(res)
                        else:
                            invalid_possibilities.append(res)

                if found_classification:
                    break

        if not found_classification and not warnings:
            warnings.append("Unable find a classification")

        return ValidationSummary(
            valid_results=valid_possibilities,
            invalid_results=invalid_possibilities,
            warnings=warnings
        )



"""A module for testing HGVS tokenization."""
import unittest

from variation.tokenizers import HGVS
from .tokenizer_base import TokenizerBase


class TestHGVSTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing HGVS Tokenization."""

    def tokenizer_instance(self):
        """Return HGVS instance."""
        return HGVS()

    def token_type(self):
        """Return HGVS token type."""
        return "HGVS"

    def fixture_name(self):
        """Return fixture name for HGVS."""
        return "hgvs"



"""A module for Genomic Deletion Tokenization."""
from typing import Dict, Optional

from variation.schemas.token_response_schema import GenomicDeletionToken
from variation.tokenizers.deletion_base import DeletionBase


class GenomicDeletion(DeletionBase):
    """Class for tokenizing Deletion at the genomic reference sequence."""

    def return_token(self, params: Dict) -> Optional[GenomicDeletionToken]:
        """Return Genomic Deletion token."""
        if self.parts["coordinate_type"] == "g":
            return GenomicDeletionToken(**params)



"""Module for testing Coding DNA DelIns Validator."""
import unittest

from variation.validators import CodingDNADelIns
from variation.classifiers import CodingDNADelInsClassifier
from .validator_base import ValidatorBase


class TestCodingDNADelInsValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Coding DNA DelIns Validator."""

    def validator_instance(self):
        """Return coding DNA delins instance."""
        return CodingDNADelIns(*self.params)

    def classifier_instance(self):
        """Return the coding DNA delins classifier instance."""
        return CodingDNADelInsClassifier()

    def fixture_name(self):
        """Return the fixture name for coding DNA delins."""
        return "coding_dna_delins"



"""A module for Protein Substitution Tokenization."""
from typing import List, Optional

from variation.schemas.token_response_schema import ProteinSubstitutionToken,\
    TokenMatchType
from .polypeptide_sequence_variation_base import PolypeptideSequenceVariationBase


class ProteinSubstitution(PolypeptideSequenceVariationBase):
    """Class for tokenizing Protein Substitution."""

    def match(self, input_string: str) -> Optional[ProteinSubstitutionToken]:
        """Return a ProteinSubstitutionToken match if one exists.

        :param str input_string: The input string to match
        :return: A ProteinSubstitutionToken if a match exists.
            Otherwise, None.
        """
        if input_string is None:
            return None

        input_string = str(input_string).lower()

        psub_parts = None
        self.psub = {
            "amino_acid": None,
            "position": None,
            "new_amino_acid": None
        }

        if "." in input_string:
            if not input_string.startswith("p."):
                return None
            p_count = input_string.count("p.")
            if p_count == 1:
                psub_parts = self.splitter.split(input_string)
            elif p_count == 2:
                psub_parts = input_string.split()
        else:
            psub_parts = self.splitter.split(input_string)

        self._get_psub(psub_parts)

        if None not in self.psub.values():
            amino_acids = {self.psub["amino_acid"],
                           self.psub["new_amino_acid"]}

            if not self._is_valid_amino_acid(amino_acids):
                return None

            return ProteinSubstitutionToken(
                token=input_string,
                input_string=input_string,
                match_type=TokenMatchType.UNSPECIFIED.value,
                ref_protein=self.psub["amino_acid"],
                alt_protein=self.psub["new_amino_acid"],
                position=self.psub["position"]
            )

        return None

    def _get_psub(self, psub_parts: List) -> None:
        """Get amino acid substitution tokens.

        :param List psub_parts: The split input string
        """
        psub_parts_len = len(psub_parts)
        if psub_parts_len == 3:
            if "p." in psub_parts[0]:
                psub_parts[0] = psub_parts[0].split("p.")[-1]
            else:
                if not psub_parts[0] and psub_parts[1] and not psub_parts[2]:
                    return

            if "(" in psub_parts[0] and ")" in psub_parts[2]:
                psub_parts[0] = psub_parts[0].split("(")[-1]
                psub_parts[2] = psub_parts[2].split(")")[0]

            self._set_psub(psub_parts[0], psub_parts[1], psub_parts[2])



"""A module for testing Genomic DelIns tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache, NucleotideCache
from variation.tokenizers import GenomicDelIns
from .tokenizer_base import TokenizerBase


class TestGenomicDelInsTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Genomic DelIns Tokenization."""

    def tokenizer_instance(self):
        """Return Genomic DelIns instance."""
        return GenomicDelIns(AminoAcidCache(), NucleotideCache())

    def token_type(self):
        """Return genomic delins token type."""
        return "GenomicDelIns"

    def fixture_name(self):
        """Return the fixture name for genomic delins."""
        return "genomic_delins"



"""A module for tokenizing genomic deletion ranges."""
from typing import Dict, Optional, List
from abc import abstractmethod

from variation.schemas.token_response_schema import TokenMatchType, \
    DeletionRange
from .tokenizer import Tokenizer


class DeletionRangeBase(Tokenizer):
    """The tokenizer class for genomic deletion range."""

    def __init__(self) -> None:
        """Initialize the Genomic Deletion Range Class."""
        self.parts = None

    def match(self, input_string: str) -> Optional[DeletionRange]:
        """Return tokens that match the input string.

        :param str input_string: Input string
        :return: DeletionRange token if a match is found
        """
        if input_string is None:
            return None

        self.parts = {
            "token": input_string,
            "input_string": input_string,
            "match_type": TokenMatchType.UNSPECIFIED.value,
            "start_pos1_del": None,
            "start_pos2_del": None,
            "end_pos1_del": None,
            "end_pos2_del": None,
            "coordinate_type": None
        }

        input_string = str(input_string).lower()
        if not input_string.endswith("del"):
            return None

        if input_string.startswith("g."):
            self.parts["coordinate_type"] = "g"
        elif input_string.startswith("c."):
            self.parts["coordinate_type"] = "c"
        elif input_string.startswith("p."):
            self.parts["coordinate_type"] = "p"
        else:
            return None

        parts = input_string.split("_")
        self._get_parts(parts)
        return self.return_token(self.parts)

    @abstractmethod
    def _get_parts(self, parts: List) -> None:
        """Set `self.parts` for genomic deletion range

        :param List parts: Parts of input string
        """
        raise NotImplementedError

    @abstractmethod
    def return_token(self, params: Dict[str, str]) -> Optional[DeletionRange]:
        """Return token instance.

        :param Dict params: Params for DeletionRange token
        :return: DeletionRange token
        """
        raise NotImplementedError



"""A module for testing Coding DNA Deletion Tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache, NucleotideCache
from variation.tokenizers import CodingDNADeletion
from .tokenizer_base import TokenizerBase


class TestCodingDNADeletionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Coding DNA Deletion Tokenization."""

    def tokenizer_instance(self):
        """Return Coding DNA Deletion instance."""
        return CodingDNADeletion(AminoAcidCache(), NucleotideCache())

    def token_type(self):
        """Return Coding DNA deletion token type."""
        return "CodingDNADeletion"

    def fixture_name(self):
        """Return the fixture name for Coding DNA Deletion."""
        return "coding_dna_deletion"



"""A module for testing Protein Substitution tokenization."""
import unittest

from variation.tokenizers.caches import AminoAcidCache
from variation.tokenizers import ProteinSubstitution
from .tokenizer_base import TokenizerBase


class TestProteinSubstitutionTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Protein Substitution Tokenization."""

    def tokenizer_instance(self):
        """Return protein substitution instance."""
        return ProteinSubstitution(AminoAcidCache())

    def token_type(self):
        """Return protein substitution token type."""
        return "ProteinSubstitution"

    def fixture_name(self):
        """Return the fixture name for protein substitution."""
        return "protein_substitution"



"""Module for testing Genomic DelIns Translator."""
import unittest

from variation.classifiers import GenomicDelInsClassifier
from variation.translators import GenomicDelIns
from variation.validators import GenomicDelIns as GENOMICDELINS_V
from .translator_base import TranslatorBase


class TestGenomicDelInsTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the Genomic DelIns Translator."""

    def classifier_instance(self):
        """Return Genomic delins instance."""
        return GenomicDelInsClassifier()

    def validator_instance(self):
        """Return genomic delins instance."""
        return GENOMICDELINS_V(*self.params)

    def translator_instance(self):
        """Return genomic delins instance."""
        return GenomicDelIns()

    def fixture_name(self):
        """Return the fixture name for genomic delins."""
        return "genomic_delins"



"""Module for testing Coding DNA DelIns Classifier."""
import unittest

from variation.classifiers import CodingDNADelInsClassifier
from .classifier_base import ClassifierBase


class TestCodingDNADelInsClassifier(ClassifierBase, unittest.TestCase):
    """A class to test the Coding DNA DelIns Classifier."""

    def classifier_instance(self):
        """Return CodingDNADelInsClassifier instance."""
        return CodingDNADelInsClassifier()

    def fixture_name(self):
        """Return CodingDNADelInsClassifier fixture name."""
        return "coding_dna_delins"



"""Module for testing Genomic Uncertain Deletion Translator."""
import unittest

from variation.classifiers import GenomicUncertainDeletionClassifier
from variation.translators import GenomicUncertainDeletion
from variation.validators import GenomicUncertainDeletion as GUD_V
from .translator_base import TranslatorBase


class TestGenomicUncertainDeletionTranslator(TranslatorBase,
                                             unittest.TestCase):
    """A class to test the Genomic Uncertain Deletion Translator."""

    def classifier_instance(self):
        """Return genomic uncertain deletion instance."""
        return GenomicUncertainDeletionClassifier()

    def validator_instance(self):
        """Return genomic uncertain deletion instance."""
        return GUD_V(*self.params)

    def translator_instance(self):
        """Return genomic uncertain deletion instance."""
        return GenomicUncertainDeletion()

    def fixture_name(self):
        """Return the fixture name for genomic uncertain deletion."""
        return "genomic_uncertain_deletion"



"""A module for testing Coding DNA Silent Mutation tokenization."""
import unittest

from variation.tokenizers import CodingDNASilentMutation
from .tokenizer_base import TokenizerBase


class TestCodingDNASilentMutationTokenizer(TokenizerBase, unittest.TestCase):
    """A class for testing Coding DNA Silent Mutation Tokenization."""

    def tokenizer_instance(self):
        """Return Coding DNA Silent Mutation instance."""
        return CodingDNASilentMutation()

    def token_type(self):
        """Return coding DNA silent mutation token type."""
        return "CodingDNASilentMutation"

    def fixture_name(self):
        """Return the fixture name for coding DNA silent mutation."""
        return "coding_dna_silent_mutation"



"""A module for the Genomic Duplication Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class GenomicDuplicationClassifier(SetBasedClassifier):
    """The Genomic Duplication Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Genomic Duplication classification type."""
        return ClassificationType.GENOMIC_DUPLICATION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["Chromosome", "GenomicDuplication"],
            ["GenomicDuplication", "GeneSymbol"],
            ["GeneSymbol", "GenomicDuplication"],
            ["HGVS", "GenomicDuplication"],
            ["ReferenceSequence", "GenomicDuplication"],
            ["LocusReferenceGenomic", "GenomicDuplication"],
            ["GenomicDuplicationRange", "GeneSymbol"],
            ["GeneSymbol", "GenomicDuplicationRange"],
            ["HGVS", "GenomicDuplicationRange"],
            ["ReferenceSequence", "GenomicDuplicationRange"],
            ["LocusReferenceGenomic", "GenomicDuplicationRange"]
        ]



"""Module for Coding DNA Silent Mutation Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import CodingDNASilentMutationToken, Token


class CodingDNASilentMutation(Translator):
    """The Coding DNA Silent Mutation Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Coding DNA Silent Mutation."""
        return type == ClassificationType.CODING_DNA_SILENT_MUTATION

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Coding DNA Silent Mutation token
        instance.
        """
        return isinstance(token, CodingDNASilentMutationToken)



"""A module for tokenizing genomic uncertain deletion."""
from typing import Optional, Union, Dict, List

from pydantic.error_wrappers import ValidationError

from variation.schemas.token_response_schema import \
    GenomicUncertainDeletionToken
from variation.tokenizers.deletion_range_base import DeletionRangeBase


class GenomicUncertainDeletion(DeletionRangeBase):
    """The tokenizer class for genomic uncertain deletion."""

    def _get_parts(self, parts: List) -> None:
        """Set parts for genomic uncertain deletion.

        :param List parts: Parts of input string
        """
        len_parts = len(parts)
        if len_parts not in [3, 4]:
            return None

        if not parts[0].startswith("g."):
            return None

        parts[0] = parts[0][2:]
        parts[len_parts - 1] = parts[len_parts - 1][:-3]

        if len_parts == 3:
            if "(" in parts[0] and ")" in parts[1]:
                # Format is: (?_#)_#del
                for part_ix, parts_field in [
                    (0, "start_pos1_del"),
                    (1, "start_pos2_del"),
                    (2, "end_pos1_del")
                ]:
                    part_val = self._check_uncertain_or_int(
                        parts[part_ix])
                    if part_val is None:
                        return None
                    else:
                        self.parts[parts_field] = part_val
            else:
                # Format is #_(#_?)del
                for part_ix, parts_field in [
                    (0, "start_pos1_del"),
                    (1, "end_pos1_del"),
                    (2, "end_pos2_del")
                ]:
                    part_val = self._check_uncertain_or_int(
                        parts[part_ix])
                    if part_val is None:
                        return None
                    else:
                        self.parts[parts_field] = part_val
            self.parts["coordinate_type"] = "g"
        elif len_parts == 4:
            for part_ix, parts_field in [
                (0, "start_pos1_del"),
                (1, "start_pos2_del"),
                (2, "end_pos1_del"),
                (3, "end_pos2_del")
            ]:
                part_val = self._check_uncertain_or_int(parts[part_ix])
                if part_val is None:
                    return None
                else:
                    self.parts[parts_field] = part_val
            self.parts["coordinate_type"] = "g"
        return None

    def _check_uncertain_or_int(self, part: str) -> Optional[Union[str, int]]:
        part = part.replace("(", "")
        part = part.replace(")", "")
        try:
            return int(part)
        except ValueError:
            if part == "?":
                return part
        return None

    def return_token(self, params: Dict) -> Optional[GenomicUncertainDeletionToken]:
        """Return Genomic Uncertain Deletion token."""
        if params["coordinate_type"] == "g":
            conditions = (
                isinstance(params["start_pos1_del"], int),
                isinstance(params["start_pos2_del"], int),
                isinstance(params["end_pos1_del"], int),
                isinstance(params["end_pos1_del"], int)
            )
            if not all(conditions):
                try:
                    prev_val = None
                    for field in ["start_pos1_del", "start_pos2_del",
                                  "end_pos1_del", "end_pos2_del"]:
                        val = self.parts[field]
                        if val not in ["?", None]:
                            if prev_val is not None:
                                if prev_val > val:
                                    return
                            prev_val = val
                    return GenomicUncertainDeletionToken(**params)
                except ValidationError:
                    return None



"""This module provides methods for handling queries."""
from typing import Tuple, Optional, List, Union, Dict
from urllib.parse import quote
import copy
import json
from os import environ

import python_jsonschema_objects
from gene.query import QueryHandler as GeneQueryHandler
from ga4gh.vrs.dataproxy import SeqRepoDataProxy
from ga4gh.vrs.extras.translator import Translator
from ga4gh.core import sha512t24u, ga4gh_identify
from ga4gh.vrs import models
from ga4gh.vrsatile.pydantic.vrs_models import Text, Allele, AbsoluteCopyNumber, \
    Haplotype, VariationSet, RelativeCopyNumber
from ga4gh.vrsatile.pydantic.vrsatile_models import VariationDescriptor, \
    CanonicalVariation, ComplexVariation
from uta_tools import SEQREPO_DATA_PATH, TRANSCRIPT_MAPPINGS_PATH, \
    LRG_REFSEQGENE_PATH, MANE_SUMMARY_PATH, UTATools

from variation import AMINO_ACID_PATH, UTA_DB_URL
from variation.schemas.app_schemas import Endpoint
from variation.schemas.hgvs_to_copy_number_schema import VALID_RELATIVE_COPY_CLASS, \
    CopyNumberType, RelativeCopyClass
from variation.schemas.token_response_schema import Nomenclature, Token, \
    CoordinateType, SequenceOntology
from variation.schemas.validation_response_schema import ValidationSummary
from variation.to_vrs import ToVRS
from variation.vrs import VRS
from variation.normalize import Normalize
from variation.classifiers import Classify
from variation.tokenizers import Tokenize
from variation.validators import Validate
from variation.translators import Translate
from variation.data_sources import CodonTable
from variation.hgvs_dup_del_mode import HGVSDupDelMode
from variation.tokenizers import GeneSymbol
from variation.tokenizers.caches import AminoAcidCache
from variation.schemas.normalize_response_schema\
    import HGVSDupDelMode as HGVSDupDelModeEnum


class QueryHandler:
    """Class for handling queries."""

    def __init__(self,
                 dynamodb_url: str = "",
                 dynamodb_region: str = "us-east-2",
                 seqrepo_data_path: str = None,
                 amino_acids_file_path: str = AMINO_ACID_PATH,
                 transcript_file_path: str = None,
                 refseq_file_path: str = None,
                 mane_data_path: str = None,
                 uta_db_url: str = UTA_DB_URL,
                 uta_db_pwd: Optional[str] = None) -> None:
        """Initialize QueryHandler instance.
        :param str dynamodb_url: URL to gene-normalizer database source.
        :param str dynamodb_region: AWS default region for gene-normalizer.
        :param str seqrepo_data_path: Path to seqrepo data directory
        :param str amino_acids_file_path: Path to amino acids file
        :param str transcript_file_path: Path to transcript mappings file
        :param str refseq_file_path: Path to refseq gene symbol file
        :param str mane_data_path: Path to refseq mane data file
        :param str uta_db_url: URL for UTA database
        :param Optional[str] uta_db_pwd: Password for UTA database user
        """
        self.amino_acid_cache = AminoAcidCache(
            amino_acids_file_path=amino_acids_file_path
        )
        self.gene_normalizer = GeneQueryHandler(db_url=dynamodb_url,
                                                db_region=dynamodb_region)

        if not seqrepo_data_path:
            seqrepo_data_path = environ.get("SEQREPO_DATA_PATH", SEQREPO_DATA_PATH)
        if not transcript_file_path:
            transcript_file_path = environ.get("TRANSCRIPT_MAPPINGS_PATH",
                                               TRANSCRIPT_MAPPINGS_PATH)
        if not refseq_file_path:
            refseq_file_path = environ.get("LRG_REFSEQGENE_PATH", LRG_REFSEQGENE_PATH)
        if not mane_data_path:
            mane_data_path = environ.get("MANE_SUMMARY_PATH", MANE_SUMMARY_PATH)

        self.uta_tools = UTATools(seqrepo_data_path=seqrepo_data_path,
                                  transcript_file_path=transcript_file_path,
                                  lrg_refseqgene_path=refseq_file_path,
                                  mane_data_path=mane_data_path,
                                  db_url=uta_db_url, db_pwd=uta_db_pwd)
        self.seqrepo_access = self.uta_tools.seqrepo_access
        self.codon_table = CodonTable(self.amino_acid_cache)
        self.uta = self.uta_tools.uta_db
        self.dp = SeqRepoDataProxy(self.seqrepo_access.seqrepo_client)
        self.tlr = Translator(data_proxy=self.dp)
        self.hgvs_dup_del_mode = HGVSDupDelMode(self.seqrepo_access)
        self.vrs = VRS(self.dp, self.seqrepo_access)
        self.to_vrs_handler = self._init_to_vrs()
        self.normalize_handler = Normalize(
            self.seqrepo_access, self.uta, self.gene_normalizer
        )

    def _init_to_vrs(self) -> ToVRS:
        """Return toVRS instance"""
        gene_symbol = GeneSymbol(self.gene_normalizer)
        tokenizer = Tokenize(self.amino_acid_cache, gene_symbol)
        classifier = Classify()
        transcript_mappings = self.uta_tools.transcript_mappings
        mane_transcript_mappings = self.uta_tools.mane_transcript_mappings
        mane_transcript = self.uta_tools.mane_transcript
        validator = Validate(
            self.seqrepo_access, transcript_mappings, gene_symbol,
            mane_transcript, self.uta, self.dp, self.tlr,
            self.amino_acid_cache, self.gene_normalizer, self.vrs
        )
        translator = Translate()
        return ToVRS(
            tokenizer, classifier, self.seqrepo_access, transcript_mappings,
            gene_symbol, self.amino_acid_cache, self.uta,
            mane_transcript_mappings, mane_transcript, validator, translator,
            self.gene_normalizer, self.hgvs_dup_del_mode
        )

    async def to_vrs(self, q: str)\
            -> Tuple[Optional[Union[List[Allele], List[AbsoluteCopyNumber],
                                    List[Text], List[Haplotype],
                                    List[VariationSet]]],
                     Optional[List[str]]]:
        """Return a VRS-like representation of all validated variations for a query.  # noqa: E501

        :param str q: The variation to translate
        :return: Validated VRS Variations and list of warnings
        """
        validations, warnings = await self.to_vrs_handler.get_validations(q)
        translations, warnings = \
            self.to_vrs_handler.get_translations(validations, warnings)

        if not translations:
            if q and q.strip():
                text = models.Text(definition=q, type="Text")
                text._id = ga4gh_identify(text)
                translations = [Text(**text.as_dict())]
            else:
                translations = None
        return translations, warnings

    async def normalize(
            self, q: str,
            hgvs_dup_del_mode: Optional[HGVSDupDelModeEnum] = HGVSDupDelModeEnum.DEFAULT  # noqa: E501
    ) -> Optional[VariationDescriptor]:
        """Return normalized Variation Descriptor for variation.

        :param q: Variation to normalize
        :param Optional[HGVSDupDelModeEnum] hgvs_dup_del_mode:
            Must be set when querying HGVS dup/del expressions.
            Must be: `default`, `cnv`, `repeated_seq_expr`, `literal_seq_expr`.
            This parameter determines how to interpret HGVS dup/del expressions
            in VRS.
        :return: Variation Descriptor for variation
        """
        validations, warnings = await self.to_vrs_handler.get_validations(
            q, endpoint_name=Endpoint.NORMALIZE, hgvs_dup_del_mode=hgvs_dup_del_mode
        )
        if not validations:
            self.normalize_handler.warnings = warnings
            return None
        return self.normalize_handler.normalize(q, validations,
                                                warnings)

    async def _get_gnomad_vcf_validations(
            self, q: str, warnings: List) -> Optional[ValidationSummary]:
        """Get gnomad vcf validation summary

        :param str q: Input query
        :param List warnings: List of warnings
        :return: ValidationSummary for a gnomad VCF query
        """
        tokens = self.to_vrs_handler.tokenizer.perform(q.strip(), warnings)
        for t in tokens:
            if t.nomenclature != Nomenclature.GNOMAD_VCF:
                warnings.append(f"{q} is not a supported gnomad vcf query")
                return None
        classifications = self.to_vrs_handler.classifier.perform(tokens)
        validation_summary = await self.to_vrs_handler.validator.perform(
            classifications, Endpoint.NORMALIZE, warnings,
            hgvs_dup_del_mode=HGVSDupDelModeEnum.LITERAL_SEQ_EXPR
        )
        if not validation_summary:
            warnings.append(f"{q} is not a valid gnomad vcf query")
            return None
        else:
            return validation_summary

    def _get_refseq_alt_ac_from_variation(self, variation: Dict) -> str:
        """Get genomic ac from variation sequence_id

        :param Dict variation: VRS variation object
        :return: RefSeq genomic accession
        """
        # genomic ac should always be in 38
        alt_ac = variation["location"]["sequence_id"]
        aliases = self.seqrepo_access.seqrepo_client.translate_identifier(
            alt_ac, target_namespaces="refseq")
        return aliases[0].split("refseq:")[-1]

    def _update_gnomad_vcf_mane_c_pos(
            self, reading_frame: int, mane_c_ac: str,
            mane_c_pos_change: Tuple[int, int], coding_start_site: int,
            warnings: List) -> Optional[Tuple[int, int]]:
        """Return updated mane c position change for a gnomad vcf variation
        depending on reading frame base

        :param int reading_frame: reading frame base
        :param str mane_c_ac: Mane transcript accession
        :param Tuple[int, int] mane_c_pos_change: Mane transcript position
            change
        :param int coding_start_site: Coding start site
        :param List warnings: List of warnings
        :return: Mane c pos start and end
        """
        if reading_frame == 1:
            # first pos
            mane_c_pos_change = \
                mane_c_pos_change[0], mane_c_pos_change[0] + 2
        elif reading_frame == 2:
            # middle pos
            mane_c_pos_change = \
                mane_c_pos_change[0] - 1, mane_c_pos_change[0] + 1
        elif reading_frame == 3:
            # last pos
            mane_c_pos_change = \
                mane_c_pos_change[0] - 2, mane_c_pos_change[0]

        if not self.to_vrs_handler.mane_transcript._validate_index(
                mane_c_ac, mane_c_pos_change, coding_start_site):
            warnings.append(
                f"{mane_c_pos_change} are not valid positions on "
                f"{mane_c_ac} with coding start site "
                f"{coding_start_site}")
            return None
        return mane_c_pos_change

    def _get_gnomad_vcf_protein_alt(
            self, classification_token: Token, reading_frame: int, strand: str,
            alt_ac: str, g_start_pos: int, g_end_pos: int) -> Optional[str]:
        """Return protein alteration that corresponds to gnomad VCF alteration

        :param Token classification_token: Classification token for query
        :param int reading_frame: cDNA reading frame number (1, 2, 3)
        :param str strand: Strand for query
        :param str alt_ac: RefSeq genomic accession
        :param int g_start_pos: Genomic start position
        :param int g_end_pos: Genomic end position
        :return: Amino acid alteration (using 1-letter codes)
        """
        alt = None
        classification_token.coordinate_type = CoordinateType.PROTEIN
        classification_token.molecule_context = "protein"
        if classification_token.alt_type in ["substitution",
                                             "silent_mutation"]:
            if classification_token.alt_type == "substitution":
                alt_nuc = classification_token.new_nucleotide
                classification_token.so_id = \
                    SequenceOntology.PROTEIN_SUBSTITUTION
            else:
                alt_nuc = classification_token.ref_nucleotide
                classification_token.so_id = SequenceOntology.SILENT_MUTATION

            ref = None
            if reading_frame == 1:
                # first pos
                if strand == "-":
                    ref, _ = self.seqrepo_access.get_reference_sequence(
                        alt_ac, g_start_pos - 1, g_end_pos + 2)
                    alt = alt_nuc + ref[1] + ref[0]
                else:
                    ref, _ = self.seqrepo_access.get_reference_sequence(
                        alt_ac, g_start_pos + 1, g_end_pos + 4)
                    alt = alt_nuc + ref[1] + ref[2]
            elif reading_frame == 2:
                # middle pos
                ref, _ = self.seqrepo_access.get_reference_sequence(
                    alt_ac, g_start_pos, g_end_pos + 3)
                alt = ref[0] + alt_nuc + ref[2]
            elif reading_frame == 3:
                # last pos
                if strand == "-":
                    ref, _ = self.seqrepo_access.get_reference_sequence(
                        alt_ac, g_start_pos + 1, g_end_pos + 4)
                    alt = ref[2] + ref[1] + alt_nuc
                else:
                    ref, _ = self.seqrepo_access.get_reference_sequence(
                        alt_ac, g_start_pos - 1, g_end_pos + 2)
                    alt = ref[0] + ref[1] + alt_nuc

            if alt and strand == "-":
                alt = self.codon_table.dna_to_rna(alt)
            else:
                alt = alt.replace("T", "U")
        elif classification_token.alt_type == "deletion":
            # There is no alt for a deletion
            classification_token.so_id = SequenceOntology.PROTEIN_DELETION
        elif classification_token.alt_type == "insertion":
            classification_token.so_id = SequenceOntology.PROTEIN_INSERTION
            alt = classification_token.inserted_sequence.replace("T", "U")
            if strand == "-":
                alt = alt[::-1]
        else:
            return None

        if alt is None:
            return None
        else:
            if len(alt) % 3 != 0:
                return None

            aa_alt = ""
            for i in range(int(len(alt) / 3)):
                aa_alt += self.codon_table.table[alt[3 * i:(3 * i) + 3]]
            return aa_alt

    async def gnomad_vcf_to_protein(
            self, q: str) -> Tuple[Optional[VariationDescriptor], List]:
        """Get protein consequence for gnomad vcf

        :param str q: gnomad vcf (chr-pos-ref-alt)
        :return: Variation Descriptor, list of warnings
        """
        q = q.strip()
        if not q:
            return self.normalize_handler.no_variation_entered()
        _id = f"normalize.variation:{quote(' '.join(q.split()))}"
        warnings = list()
        valid_list = list()
        validations = await self._get_gnomad_vcf_validations(q, warnings)
        if not validations:
            return self.normalize_handler.text_variation_resp(q, _id, warnings)

        all_warnings = list()
        for valid_result in validations.valid_results:
            warnings = list()
            # all gnomad vcf will be alleles with a literal seq expression
            variation = valid_result.variation
            classification_token = valid_result.classification_token
            alt_ac = self._get_refseq_alt_ac_from_variation(variation)

            # 0-based
            g_start_pos = None
            g_end_pos = None
            if classification_token.alt_type == "deletion":
                g_start_pos = classification_token.start_pos_del
                g_end_pos = classification_token.end_pos_del
            elif classification_token.alt_type == "insertion":
                g_start_pos = classification_token.start_pos_flank
                g_end_pos = classification_token.end_pos_flank
            elif classification_token.alt_type in ["silent_mutation",
                                                   "substitution"]:
                g_start_pos = classification_token.position
                g_end_pos = classification_token.position
                ref_seq, w = self.seqrepo_access.get_reference_sequence(
                    alt_ac, g_start_pos)
                if not ref_seq:
                    all_warnings.append(w)
                else:
                    if ref_seq != classification_token.ref_nucleotide:
                        all_warnings.append(
                            f"Expected {classification_token.ref_nucleotide}"
                            f" but found {ref_seq} on {alt_ac} at position"
                            f" {g_start_pos}"
                        )
                        continue
            else:
                all_warnings.append(
                    f"{classification_token.alt_type} alt_type not supported"
                )
                continue

            g_start_pos -= 1
            g_end_pos -= 1

            transcripts = await self.uta.get_transcripts_from_genomic_pos(
                alt_ac, g_start_pos)

            if not transcripts:
                all_warnings.append(f"Unable to get transcripts given {alt_ac}"
                                    f" and {g_start_pos + 1}")
                continue
            mane_data = self.to_vrs_handler.mane_transcript_mappings.get_mane_from_transcripts(transcripts)  # noqa: E501
            mane_data_len = len(mane_data)
            for i in range(mane_data_len):
                index = mane_data_len - i - 1
                current_mane_data = mane_data[index]
                mane_c_ac = current_mane_data["RefSeq_nuc"]
                mane_tx_genomic_data = await self.uta.get_mane_c_genomic_data(
                    mane_c_ac, alt_ac, g_start_pos, g_end_pos)
                if not mane_tx_genomic_data:
                    all_warnings.append("Unable to get mane transcript and "
                                        "genomic data")

                coding_start_site = mane_tx_genomic_data["coding_start_site"]
                mane_c_pos_change = \
                    self.to_vrs_handler.mane_transcript.get_mane_c_pos_change(
                        mane_tx_genomic_data, coding_start_site)

                # We use 1-based
                reading_frame = self.to_vrs_handler.mane_transcript.get_reading_frame(
                    mane_c_pos_change[0] + 1)
                if classification_token.alt_type in ["substitution",
                                                     "silent_mutation"]:
                    mane_c_pos_change = self._update_gnomad_vcf_mane_c_pos(
                        reading_frame, mane_c_ac, mane_c_pos_change,
                        coding_start_site, warnings)
                    if mane_c_pos_change is None:
                        if len(warnings) > 0:
                            all_warnings.append(warnings[0])
                        continue

                mane_p = self.to_vrs_handler.mane_transcript._get_mane_p(
                    current_mane_data,
                    (mane_c_pos_change[0] + 1, mane_c_pos_change[1] + 1)
                )
                if mane_p["pos"][0] > mane_p["pos"][1]:
                    mane_p["pos"] = (mane_p["pos"][1], mane_p["pos"][0])
                p_ac = mane_p["refseq"]
                valid_result.identifier = p_ac
                aa_alt = self._get_gnomad_vcf_protein_alt(
                    classification_token, reading_frame,
                    mane_tx_genomic_data["strand"], alt_ac,
                    g_start_pos, g_end_pos)
                if aa_alt or classification_token.alt_type == "deletion":
                    variation = self.vrs.to_vrs_allele(
                        p_ac, mane_p["pos"][0], mane_p["pos"][1], "p",
                        classification_token.alt_type, [], alt=aa_alt
                    )
                    if variation:
                        valid_list.append(
                            self.normalize_handler.get_variation_descriptor(
                                q, variation, valid_result, _id, warnings,
                                gene=current_mane_data["HGNC_ID"]))

        if valid_list:
            return valid_list[0]
        else:
            if all_warnings:
                return self.normalize_handler.text_variation_resp(
                    q, _id, all_warnings)
            else:
                return self.normalize_handler.text_variation_resp(
                    q, _id, [f"Unable to get protein variation for {q}"])

    def canonical_spdi_to_categorical_variation(
            self, q: str, complement: bool = False
    ) -> Tuple[Optional[Union[CanonicalVariation, ComplexVariation]], List]:
        """Return categorical variation for canonical SPDI

        :param str q: Canonical SPDI
        :param bool complement: This field indicates that a categorical
            variation is defined to include (false) or exclude (true) variation
             concepts matching the categorical variation. This is equivalent to a
             logical NOT operation on the categorical variation properties.
        :return: Tuple containing CanonicalVariation and list of warnings
        """
        q = q.strip()
        if not q:
            return self.normalize_handler.no_variation_entered()
        warnings = list()

        variation = None
        try:
            variation = self.tlr.translate_from(q, fmt="spdi")
        except (ValueError, python_jsonschema_objects.validators.ValidationError) as e:
            warnings.append(f"vrs-python translator raised error: {e}")
        except KeyError as e:
            warnings.append(f"vrs-python translator raised error: "
                            f"seqrepo could not translate identifier {e}")
        if warnings or not variation:
            return None, warnings

        spdi_parts = q.split(":")
        ac = spdi_parts[0]
        start_pos = int(spdi_parts[1])  # inter-residue, 0 based
        deleted_seq = spdi_parts[2]
        end_pos = start_pos + len(deleted_seq)
        try:
            sequence = self.seqrepo_access.seqrepo_client.fetch(
                ac, start_pos, end=end_pos)
        except ValueError as e:
            warnings.append(str(e))
        else:
            if not sequence:
                warnings.append(f"Position, {start_pos}, does not exist on {ac}")
            else:
                if deleted_seq != sequence:
                    warnings.append(f"Expected to find reference sequence"
                                    f" {deleted_seq} but found {sequence} on {ac}")
        if warnings:
            return None, warnings

        variation.location._id = ga4gh_identify(variation.location)
        canonical_variation = {
            "type": "CanonicalVariation",
            "complement": complement,
            "variation": variation.as_dict()
        }

        cpy_canonical_variation = copy.deepcopy(canonical_variation)
        cpy_canonical_variation["variation"] = canonical_variation["variation"]["_id"].split(".")[-1]  # noqa: E501
        serialized = json.dumps(
            cpy_canonical_variation, sort_keys=True, separators=(",", ":"), indent=None
        ).encode("utf-8")
        digest = sha512t24u(serialized)
        # VCC = variation categorical canonical
        canonical_variation["_id"] = f"ga4gh:VCC.{digest}"
        canonical_variation = CanonicalVariation(**canonical_variation)
        return canonical_variation, warnings

    def _hgvs_to_cnv_resp(
        self, copy_number_type: CopyNumberType, hgvs_expr: str, do_liftover: bool,
        validations: Tuple[Optional[ValidationSummary], Optional[List[str]]],
        warnings: List[str]
    ) -> Tuple[Optional[Union[AbsoluteCopyNumber, RelativeCopyNumber, Text]], List[str]]:  # noqa: E501
        """Return copy number variation and warnings response

        :param CopyNumberType copy_number_type: The type of copy number variation
        :param str hgvs_expr: HGVS expression
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        :param Tuple[Optional[ValidationSummary], Optional[List[str]]]: Validation
            summary and warnings for hgvs_expr
        :param List[str] warnings: List of warnings
        :return: CopyNumberVariation and warnings
        """
        variation = None
        if do_liftover:
            valid_result = self.normalize_handler.get_valid_result(
                hgvs_expr, validations, [])
            if valid_result:
                variation = valid_result.variation
            else:
                warnings.append(f"Unable to translate {hgvs_expr} to "
                                f"copy number variation")
        else:
            translations, warnings = \
                self.to_vrs_handler.get_translations(validations, warnings)
            if translations:
                variation = translations[0]

        if not variation:
            if hgvs_expr and hgvs_expr.strip():
                text = models.Text(definition=hgvs_expr, type="Text")
                text._id = ga4gh_identify(text)
                variation = Text(**text.as_dict())
        else:
            if copy_number_type == CopyNumberType.ABSOLUTE:
                variation = AbsoluteCopyNumber(**variation)
            else:
                variation = RelativeCopyNumber(**variation)
        return variation, warnings

    async def hgvs_to_absolute_copy_number(
        self, hgvs_expr: str, baseline_copies: Optional[int] = None,
        do_liftover: bool = False
    ) -> Tuple[Optional[AbsoluteCopyNumber], List[str]]:
        """Given hgvs, return abolute copy number variation

        :param str hgvs_expr: HGVS expression
        :param Optional[int] baseline_copies: Baseline copies number
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        :return: Absolute Copy Number Variation and warnings
        """
        validations, warnings = await self.to_vrs_handler.get_validations(
            hgvs_expr, endpoint_name=Endpoint.HGVS_TO_ABSOLUTE_CN,
            hgvs_dup_del_mode=CopyNumberType.ABSOLUTE,
            baseline_copies=baseline_copies, do_liftover=do_liftover
        )
        return self._hgvs_to_cnv_resp(
            CopyNumberType.ABSOLUTE, hgvs_expr, do_liftover, validations, warnings)

    async def hgvs_to_relative_copy_number(
        self, hgvs_expr: str, relative_copy_class: RelativeCopyClass,
        do_liftover: bool = False
    ) -> Tuple[Optional[RelativeCopyNumber], List[str]]:
        """Given hgvs, return relative copy number variation

        :param str hgvs_expr: HGVS expression
        :param RelativeCopyClass relative_copy_class: The relative copy class
        :param bool do_liftover: Whether or not to liftover to GRCh38 assembly
        :return: Relative Copy Number Variation and warnings
        """
        if relative_copy_class and relative_copy_class.lower() not in VALID_RELATIVE_COPY_CLASS:  # noqa: E501
            return None, [f"{relative_copy_class} is not a valid relative copy class: "
                          f"{VALID_RELATIVE_COPY_CLASS}"]

        validations, warnings = await self.to_vrs_handler.get_validations(
            hgvs_expr, endpoint_name=Endpoint.HGVS_TO_RELATIVE_CN,
            hgvs_dup_del_mode=CopyNumberType.RELATIVE,
            relative_copy_class=relative_copy_class, do_liftover=do_liftover
        )

        return self._hgvs_to_cnv_resp(
            CopyNumberType.RELATIVE, hgvs_expr, do_liftover, validations, warnings)



"""A module for Duplication Tokenization Base Class."""
from abc import abstractmethod
from typing import Optional, List

from variation.schemas.token_response_schema import Duplication, \
    TokenMatchType, DuplicationAltType, Token
from .tokenizer import Tokenizer


class DuplicationBase(Tokenizer):
    """Class for tokenizing Deletions."""

    def __init__(self) -> None:
        """Initialize the Deletion Base Class."""
        self.parts = None

    def match(self, input_string: str) -> Optional[Duplication]:
        """Return tokens that match the input string."""
        if input_string is None:
            return None

        self.parts = {
            "token": input_string,
            "input_string": input_string,
            "match_type": TokenMatchType.UNSPECIFIED.value,
            "start_pos1_dup": None,
            "start_pos2_dup": None,
            "end_pos1_dup": None,
            "end_pos2_dup": None,
            "coordinate_type": None,
            "alt_type": DuplicationAltType.DUPLICATION
        }

        input_string = str(input_string).lower()
        if not input_string.endswith("dup"):
            return None

        parts = [input_string[:-3]]
        self._get_parts(parts)
        return self.return_token()

    @abstractmethod
    def _get_parts(self, parts: List) -> None:
        """Get parts for DelIns

        :param List parts: Parts of input string
        """
        raise NotImplementedError

    @abstractmethod
    def return_token(self) -> Optional[Token]:
        """Return token instance."""
        raise NotImplementedError



"""Module for testing Genomic Silent Mutation Validator."""
import unittest

from variation.validators import GenomicSilentMutation
from variation.classifiers import GenomicSilentMutationClassifier
from .validator_base import ValidatorBase


class TestGenomicSilentMutationValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Genomic Silent Mutation Validator."""

    def validator_instance(self):
        """Return genomic silent mutation instance."""
        return GenomicSilentMutation(*self.params)

    def classifier_instance(self):
        """Return the genomic silent mutation classifier instance."""
        return GenomicSilentMutationClassifier()

    def fixture_name(self):
        """Return the fixture name for genomic silent mutation."""
        return "genomic_silent_mutation"



"""A module for the Coding DNA Substitution Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class CodingDNASubstitutionClassifier(SetBasedClassifier):
    """The Coding DNA Substitution Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Coding DNA Substitution classification type."""
        return ClassificationType.CODING_DNA_SUBSTITUTION

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["CodingDNASubstitution"],
            ["GeneSymbol", "ProteinSubstitution", "CodingDNASubstitution"],
            ["CodingDNASubstitution", "GeneSymbol"],
            ["GeneSymbol", "CodingDNASubstitution"],
            ["HGVS", "CodingDNASubstitution"],
            ["ReferenceSequence", "CodingDNASubstitution"],
            ["LocusReferenceGenomic", "CodingDNASubstitution"]
        ]



"""Module for Genomic Deletion Range Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import \
    GenomicDeletionRangeToken, Token


class GenomicDeletionRange(Translator):
    """The Genomic Insertion Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Genomic Insertion."""
        return type == ClassificationType.GENOMIC_DELETION_RANGE

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Genomic Deletion Range token instance."""
        return isinstance(token, GenomicDeletionRangeToken)



"""Module for testing Protein Deletion Validator."""
import unittest

from variation.validators import ProteinDeletion
from variation.classifiers import ProteinDeletionClassifier
from .validator_base import ValidatorBase


class TestProteinDeletionValidator(ValidatorBase, unittest.TestCase):
    """A class to test the Protein Deletion Validator."""

    def validator_instance(self):
        """Return protein deletion instance."""
        return ProteinDeletion(*self.aa_params)

    def classifier_instance(self):
        """Return the protein deletion classifier instance."""
        return ProteinDeletionClassifier()

    def fixture_name(self):
        """Return the fixture name for protein deletion."""
        return "protein_deletion"



"""A module for Genomic Substitution Tokenization."""
from typing import Dict, Optional

from variation.schemas.token_response_schema import GenomicSubstitutionToken
from .single_nucleotide_variation_base import SingleNucleotideVariationBase


class GenomicSubstitution(SingleNucleotideVariationBase):
    """Class for tokenizing SNV Substitution at the linear genomic
    reference sequence.
    """

    def return_token(self, params: Dict) -> Optional[GenomicSubstitutionToken]:
        """Return Genomic Substitution token."""
        if self.sub["coordinate_type"] == "g" and \
                self.sub["ref_nucleotide"] is not None and \
                self.sub["new_nucleotide"] != "=":
            return GenomicSubstitutionToken(**params)



"""A module for testing tokenizer classes."""
import yaml

from tests import PROJECT_ROOT


class TokenizerBase:
    """The tokenizer base class."""

    def setUp(self):
        """Set up the test test cases."""
        with open(f"{PROJECT_ROOT}/tests/fixtures/tokenizers.yml") as stream:
            self.all_fixtures = yaml.safe_load(stream)
        self.fixtures = self.all_fixtures.get(
            self.fixture_name(),
            {"should_match": [], "should_not_match": []}
        )

    def tokenizer_instance(self):
        """Check that tokenizer_instance method is implemented."""
        raise NotImplementedError()

    def token_type(self):
        """Check that token_type method is implemented."""
        raise NotImplementedError()

    def fixture_name(self):
        """Check that fixture_name method is implemented."""
        raise NotImplementedError()

    def test_matches(self):
        """Test that tokenizer matches correctly."""
        for x in self.fixtures["should_match"]:
            res = self.tokenizer_instance().match(x["token"])
            self.assertIsNotNone(res, msg=x)
            self.assertEqual(self.token_type(), res.token_type)

    def test_not_matches(self):
        """Test that tokenizer matches correctly."""
        for x in self.fixtures["should_not_match"]:
            res = self.tokenizer_instance().match(x["token"])
            try:
                self.assertIsNone(res, msg=x)
            except AssertionError:
                assert self.token_type() != res.token_type



"""A module for Polypeptide Sequence Variation Tokenization Base Class."""
import re
from abc import abstractmethod
from typing import List, Optional

from .caches import AminoAcidCache
from .tokenizer import Tokenizer
from ..schemas.token_response_schema import Token


class PolypeptideSequenceVariationBase(Tokenizer):
    """Class for tokenizing Polypeptide Sequence Variations."""

    def __init__(self, amino_acid_cache: AminoAcidCache) -> None:
        """Initialize the Polypeptide Sequence Variation Base Class.

        :param AminoAcidCache amino_acid_cache: Valid amino acid codes.
        """
        self.amino_acid_cache = amino_acid_cache
        self.splitter = re.compile(r"(\d+)")
        self.psub = None

    def _set_psub(self, amino_acid: str, position: int, new_amino_acid: str) -> None:
        """Initialize protein substitution.

        :param str amino_acid: Reference amino acid
        :param int position: The position of the amino acid substituted
        :param str new_amino_acid: The new amino_acid
        """
        self.psub["amino_acid"] = amino_acid.upper() if len(amino_acid) == 1 \
            else self.amino_acid_cache.convert_three_to_one(amino_acid)
        self.psub["position"] = int(position)
        self.psub["new_amino_acid"] = new_amino_acid.upper() if \
            len(new_amino_acid) == 1 else \
            self.amino_acid_cache.convert_three_to_one(new_amino_acid)

    def _is_valid_amino_acid(self, amino_acids: List) -> bool:
        """Return whether or not amino acids are valid."""
        for amino_acid_code in amino_acids:
            if not self.amino_acid_cache.__contains__(amino_acid_code):
                return False
        return True

    @abstractmethod
    def match(self, input_string: str) -> Optional[Token]:
        """Return tokens that match the input string."""
        raise NotImplementedError



"""Package level import."""
from .token_response_schema import TokenResponseSchema  # noqa: F401
from .classification_response_schema import ClassificationResponseSchema  # noqa: F401, E501
from .validation_response_schema import ValidationResponseSchema  # noqa: F401
from .to_vrs_response_schema import ToVRSService  # noqa: F401, E501
from .normalize_response_schema import NormalizeService, ServiceMeta  # noqa: F401, E501



"""A module for the Coding DNA DelIns Classifier."""
from typing import List

from variation.schemas.classification_response_schema import ClassificationType
from .set_based_classifier import SetBasedClassifier


class CodingDNADelInsClassifier(SetBasedClassifier):
    """The Coding DNA DelIns Classifier class."""

    def classification_type(self) -> ClassificationType:
        """Return the Coding DNA DelIns classification type."""
        return ClassificationType.CODING_DNA_DELINS

    def exact_match_candidates(self) -> List[List[str]]:
        """Return the exact match token type candidates."""
        return [
            ["CodingDNADelIns"],
            ["GeneSymbol", "ProteinSubstitution", "CodingDNADelIns"],
            ["CodingDNADelIns", "GeneSymbol"],
            ["GeneSymbol", "CodingDNADelIns"],
            ["HGVS", "CodingDNADelIns"],
            ["ReferenceSequence", "CodingDNADelIns"],
            ["LocusReferenceGenomic", "CodingDNADelIns"]
        ]



"""Module for Polypeptide Truncation Translation."""
from variation.translators.translator import Translator
from variation.schemas.classification_response_schema import ClassificationType
from variation.schemas.token_response_schema import PolypeptideTruncationToken, Token


class PolypeptideTruncation(Translator):
    """The Polypeptide Truncation Translator class."""

    def can_translate(self, type: ClassificationType) -> bool:
        """Return if classification type is Polypeptide Truncation."""
        return type == ClassificationType.POLYPEPTIDE_TRUNCATION

    def is_token_instance(self, token: Token) -> bool:
        """Return if the token is an Polypeptide Truncation token instance."""
        return isinstance(token, PolypeptideTruncationToken)



"""Module for testing Genomic Uncertain Deletion Classifier."""
import unittest

from variation.classifiers import GenomicUncertainDeletionClassifier
from .classifier_base import ClassifierBase


class TestGenomicUncertainDeletionClassifier(ClassifierBase,
                                             unittest.TestCase):
    """A class to test the Genomic Uncertain Deletion Classifier."""

    def classifier_instance(self):
        """Return Genomic Uncertain Deletion Classifier instance."""
        return GenomicUncertainDeletionClassifier()

    def fixture_name(self):
        """Return Genomic Uncertain Deletion fixture name."""
        return "genomic_uncertain_deletion"



"""Module for testing polypeptide truncation Translator."""
import unittest

from variation.classifiers import PolypeptideTruncationClassifier
from variation.translators import PolypeptideTruncation
from variation.validators import PolypeptideTruncation as PT_V
from .translator_base import TranslatorBase


class TestPolypeptideTruncationTranslator(TranslatorBase, unittest.TestCase):
    """A class to test the polypeptide truncation Translator."""

    def classifier_instance(self):
        """Return polypeptide truncation instance."""
        return PolypeptideTruncationClassifier()

    def validator_instance(self):
        """Return polypeptide truncation instance."""
        return PT_V(*self.aa_params)

    def translator_instance(self):
        """Return polypeptide truncation instance."""
        return PolypeptideTruncation()

    def fixture_name(self):
        """Return the fixture name for polypeptide truncation."""
        return "polypeptide_truncation"



"""The module for Protein Substitution Validation."""
from variation.schemas.classification_response_schema import \
    ClassificationType
from variation.schemas.token_response_schema import Token
from .polypeptide_sequence_variation_base import PolypeptideSequenceVariationBase


class ProteinSubstitution(PolypeptideSequenceVariationBase):
    """The Protein Substitution Validator class."""

    def variation_name(self) -> str:
        """Return the variation name."""
        return "protein substitution"

    def is_token_instance(self, t: Token) -> bool:
        """Check that token is Protein Substitution."""
        return t.token_type == "ProteinSubstitution"

    def validates_classification_type(
            self, classification_type: ClassificationType) -> bool:
        """Return whether or not the classification type is protein
        substitution.
        """
        return classification_type == ClassificationType.PROTEIN_SUBSTITUTION



"""A module to cache amino acid codes."""
from typing import Set
import csv

from variation import AMINO_ACID_PATH


class AminoAcidCache:
    """A class to cache amino acid codes."""

    def __init__(self, amino_acids_file_path: str = AMINO_ACID_PATH) -> None:
        """Initialize the AminoAcidCache class."""
        self._amino_acids_file = amino_acids_file_path
        self.amino_acid_code_conversion = dict()
        self.__amino_acid_codes = self.__load_amino_acid_codes()

    def __contains__(self, item: str) -> bool:
        """Return whether a string in the amino acid cache."""
        return item.upper() in self.__amino_acid_codes

    def __load_amino_acid_codes(self) -> Set[str]:
        """Load amino acid cache with amino acid codes.

        :return: A set of valid amino acid codes.
        """
        with open(self._amino_acids_file, "r") as f:
            data = list(csv.reader(f))
            for row in data:
                self.amino_acid_code_conversion[row[2]] = row[1]

        return {item.upper() for sublist in data for item in sublist}

    def convert_three_to_one(self, three_letter_amino_acid: str) -> str:
        """Convert a 3 letter amino acid code to a 1 letter amino acid code.

        :param str three_letter_amino_acid: Amino Acid Code to convert
        :return: A str of the one letter protein code
        """
        if three_letter_amino_acid.upper() == "TER":
            return "*"
        for one_letter, three_letter in self.amino_acid_code_conversion.items():  # noqa: E501
            if three_letter.upper() == three_letter_amino_acid.upper():
                return one_letter.upper()
